#
#   Copyright (c) 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=too-few-public-methods
'''This module provides constants and functions common to both the
Django-based server and code that uses the EPIC API (without Django).
In other words, this module may not have any Django imports or
dependencies.

'''

class Enum:
    MOUNTING_SMD = 0
    MOUNTING_THD = 1
    MOUNTING_CHASSIS = 2
    MOUNTING_FREE = 3
    MOUNTING_CHOICES = (
        (MOUNTING_SMD, 'SMD'),
        (MOUNTING_THD, 'THD'),
        (MOUNTING_CHASSIS, 'chassis'),
        (MOUNTING_FREE, 'free')
    )
    STATUS_PREVIEW = 0
    STATUS_ACTIVE = 1
    STATUS_PREFERRED = 2
    STATUS_DEPRECATED = 3
    STATUS_OBSOLETE = 4
    STATUS_CHOICES = (
        (STATUS_PREVIEW, 'preview'),
        (STATUS_ACTIVE, 'active'),
        (STATUS_PREFERRED, 'preferred'),
        (STATUS_DEPRECATED, 'deprecated'),
        (STATUS_OBSOLETE, 'obsolete'),
    )
    # deprecated parts are orderable, but preview or obsolete parts
    # are not:
    STATUS_ORDERABLE = (STATUS_ACTIVE, STATUS_PREFERRED, STATUS_DEPRECATED)

    # What modified a part's BOM (assembly-items list) last.  USER
    # means that the BOM was edited by an interactive user whereas
    # TOOL means it was generated by an automated tool (e.g.,
    # kicad-to-epic-bom).
    LAST_MOD_TYPE_USER	= 0
    LAST_MOD_TYPE_TOOL	= 1
    LAST_MOD_CHOICES = (
        (LAST_MOD_TYPE_USER, 'user'),
        (LAST_MOD_TYPE_TOOL, 'tool')
    )

    TXTN_STATUS_OPEN = 0
    TXTN_STATUS_CLOSED = 1
    TXTN_STATUS_CHOICES = (
        (TXTN_STATUS_OPEN, 'open'),
        (TXTN_STATUS_CLOSED, 'closed'),
    )

def format_part_number(part):
    return 'EP%05u' % part

def strchoice(choices, val):
    for key, name in choices:
        if key == val:
            return name
    return val



#
#   Copyright (c) 2014-2019, 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=too-few-public-methods, no-member
from __future__ import print_function

import hashlib
import math
import re

from datetime import date, timedelta
from decimal import Decimal
from django.conf import settings
from django.core.validators import MinValueValidator, MaxValueValidator
from django.core.cache import caches
from django.db import models
from django.utils import timezone
from django.utils.html import escape
from django.contrib.humanize.templatetags.humanize import intcomma

from epic.compat import reverse
from epic.base import Enum, format_part_number, strchoice

# pylint: disable=arguments-differ
# pylint: disable=too-many-lines

PART_PRICE_VALIDATORS = [MinValueValidator(Decimal('0.000001'))]
CACHE = caches['epic']

def part_cost(line_cost, qty):
    if qty <= 0:
        return 0
    return Decimal('%.6f' % (float(line_cost) / qty))

def html_part_link(part):
    """Return an HTML link for PART.  PART may be either a Part object or
    a part number.
    """
    if isinstance(part, Part):
        pn = part.id
    else:
        pn = part
    name = format_part_number(pn)
    url = reverse('epic:part_detail', kwargs={'pk': pn})
    return '<a href="%s">%s</a>' % (url, name)

def html_list_or_none(l):
    """If L is empty, return 'none', otherwise, join the list using ','
    as a separator.
    """
    html = ', '.join(l)
    if html != '':
        return html
    return '<em>&mdash;None&mdash;</em>'

class Datasheet(models.Model):
    name = models.CharField(max_length=128, unique=True, blank=False)
    ds_file = models.FileField(
        upload_to=getattr(settings, 'EPIC_DATASHEET_DIR', 'epic/datasheets'),
        blank=False,
        verbose_name='Datasheet File')
    notes = models.CharField(
        max_length=1024, help_text='For misc. information that may be useful '
        'along with the datasheet, such as website retrieved from.',
        blank=True)
    md5sum = models.CharField(max_length=36, null=True)

    class Meta:
        ordering = ('name',)

    def __str__(self):
        return '%s' % self.name

    def html_link(self):
        url = reverse('epic:datasheet_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.name)

    def save(self, *args, **kwargs):
        md5 = hashlib.md5()
        for chunk in self.ds_file.chunks():
            md5.update(chunk)
        self.md5sum = md5.hexdigest()
        super().save(*args, **kwargs)

    # See signals.datasheet_delete() for deletion cleanup.

class Part(models.Model):
    val = models.CharField(max_length=31, verbose_name='Value', blank=True,
                           default='',
                           help_text='The primary value of the part '
                           'such as resistance for a resistor or '
                           'capacitance for a capacitor.')
    descr = models.CharField(max_length=127, verbose_name='Description',
                             blank=True, default='',
                             help_text='Brief listing of key parameters of '
                             'the part, such as temperature range, drift, '
                             'max. voltages, etc.')
    footprint = models.CharField(max_length=63, blank=True, default='',
                                 help_text='The part\'s PCB footprint.')
    mfg = models.CharField(max_length=31, verbose_name='Manufacturer',
                           help_text='The name of the manufacturer of the '
                           'part.')
    mfg_pn = models.CharField(max_length=31,
                              verbose_name='Manufacturer\'s Part #')
    mounting = models.IntegerField(choices=Enum.MOUNTING_CHOICES,
                                   default=Enum.MOUNTING_SMD, blank=False,
                                   verbose_name='Mounting Type',
                                   help_text='How the part is mounted to a '
                                   'PCB (free if it\'s not mounted at all, '
                                   'such as a plug).')
    target_price = models.DecimalField(max_digits=10, decimal_places=6,
                                       validators=PART_PRICE_VALIDATORS,
                                       verbose_name='Target price [$]',
                                       help_text='The expected price of the '
                                       'part.')
    overage = models.DecimalField(max_digits=6, decimal_places=3,
                                  validators=[
                                      MinValueValidator(Decimal('0')),
                                      MaxValueValidator(Decimal('100'))],
                                  verbose_name='Overage [%]',
                                  help_text='Worst-case percentage of parts '
                                  'we expect to lose due to attrition at the '
                                  'assembly-house.')
    spq = models.IntegerField(default=1,
                              verbose_name='Standard-Package Qty',
                              help_text='The number of parts in '
                              'a standard package.  For example, 10,000 '
                              'pieces in a reel.')
    lead_time = models.PositiveIntegerField(verbose_name='Lead-time [weeks]',
                                            help_text='Lead-time in weeks.')
    status = models.IntegerField(choices=Enum.STATUS_CHOICES,
                                 default=Enum.STATUS_PREVIEW, blank=False,
                                 verbose_name='Life-time Status',
                                 help_text='The life-time status of this '
                                 'part. Parts marked `preview\' and '
                                 '`obsolete\' are not considered orderable.')
    substitute = models.ForeignKey('self', null=True, blank=True,
                                   verbose_name='Substitutes',
                                   help_text='List of other parts that are '
                                   '(identical) substitutes for this part.',
                                   on_delete=models.CASCADE)
    last_bom_mod_type = models.IntegerField(choices=Enum.LAST_MOD_CHOICES,
                                            default=Enum.LAST_MOD_TYPE_USER,
                                            blank=False)
    last_bom_mod_name = models.CharField(max_length=31, blank=True, default='',
                                         help_text='Name of entity '
                                         'which last modified this part.')
    datasheet = models.ForeignKey(Datasheet, null=True, blank=True,
                                  on_delete=models.CASCADE)
    notes = models.TextField(blank=True, default='', help_text='Notes '
                             'regarding this part.')

    class Meta:
        ordering = ['id']
        unique_together = ('mfg', 'mfg_pn')
        index_together = [
            ['mfg', 'mfg_pn']
        ]

    def __str__(self):
        return '%s' % (format_part_number(self.id))

    @classmethod
    def flush_caches(cls, sender, **kwargs):
        # pylint: disable=unused-argument
        """See signals.flush_caches()."""
        # clear the caches when a part changes:
        CACHE.clear()

    def equivalent_parts(self):
        """Return the set of equivalent parts for this part.  The set includes
        the part itself.  However, if the part doesn't have an id yet
        (newly created and not saved yet), the empty set is returned.

        """

        def get_set(equiv_parts, part):
            if part.id in equiv_parts:
                return equiv_parts[part.id]
            return frozenset([part.id])

        if self.id is None:
            # newly created parts don't have a valid id yet so they can't
            # be hashed, which also means nobody can be referring to it,
            # which means its equivalent-parts-set is empty.
            return frozenset([])

        equiv_parts = CACHE.get('equiv-parts')
        if equiv_parts is None:
            equiv_parts = {}
            for p in Part.objects.exclude(substitute__isnull=True):
                union = get_set(equiv_parts, p)
                union |= get_set(equiv_parts, p.substitute)
                for q_id in union:
                    equiv_parts[q_id] = union
            CACHE.set('equiv-parts', equiv_parts, timeout=None)
        return [Part.objects.get(pk=id) for id in get_set(equiv_parts, self)]

    def best_part(self):
        """Among the set of parts equivalent to this part, find orderable,
        preferred, or lowest-cost part.  Returns part itself if nothing
        better is available.

        """
        best_id = CACHE.get('best-part-' + str(self.id))
        if best_id is not None:
            return Part.objects.get(pk=best_id)

        best_part = self		# may not be orderable...
        for p in self.equivalent_parts():
            if p == self or not p.is_orderable():
                continue

            if (p.status == Enum.STATUS_PREFERRED) != \
               (best_part.status == Enum.STATUS_PREFERRED):
                if p.status == Enum.STATUS_PREFERRED:
                    best_part = p
                continue

            target_pr = p.target_price
            best_pr = best_part.target_price
            if not best_part.is_orderable() \
               or target_pr < best_pr \
               or (target_pr == best_pr and p.id < best_part.id):
                best_part = p
        for p in self.equivalent_parts():
            CACHE.set('best-part-' + str(self.id), best_part.id, timeout=None)
        return best_part

    def best_vendor(self):
        """Return the vendor that has the lowest price for this part."""
        vps = Vendor_Part.objects.filter(part_id=self.id) \
                                 .order_by('price')
        for vp in vps:
            if vp.status in Enum.STATUS_ORDERABLE:
                return vp.vendor
        return None

    def is_orderable(self):
        return self.status in Enum.STATUS_ORDERABLE

    def assembly_items(self):
        return Assembly_Item.objects.filter(assy_id=self.id)

    def html_link(self):
        return html_part_link(self)

    def avg_cost(self):
        '''Get average cost for a part, including any of its substitutes.  The
        average is calculated across orders with shipments within the past
        365 days.  If there is no shipment within 365, the cost is
        calculated based on the most recent shipment alone.  If there are
        no shipments at all, return the part's target price.
        '''
        CACHE_ID = 'total-costs'
        COST = 0
        QTY = 1
        total = CACHE.get(CACHE_ID)
        if total is None:
            total = {}
            cut_off_ts = timezone.now() - timedelta(days=365)

            deltas = Delta.objects.filter(adj__gt=0) \
                                  .filter(txtn__shipment__ts__gte=cut_off_ts) \
                                  .order_by('txtn_id')
            for d in deltas:
                cost = d.txtn.shipment.cost_details()
                overhead_ratio = 0
                if cost['parts'] > 0:
                    overhead_ratio = cost['overhead'] / cost['parts']

                if d.part.id not in cost['part_detail']:
                    continue
                pc = cost['part_detail'][d.part.id]

                if d.part.id not in total:
                    total[d.part.id] = [0, 0]
                if d.txtn.shipment.from_warehouse is None:
                    piece_cost = float(pc['cost']) / pc['qty']
                    total[d.part.id][COST] += ((d.adj * piece_cost)
                                               * (1 + float(overhead_ratio)))
                    total[d.part.id][QTY] += d.adj
                else:
                    # For inter-warehouse-shipments, allocate the
                    # shipping costs in proportion to the
                    # target-cost of the parts
                    part_ratio = (pc['qty'] * d.part.target_price) \
                                 / cost['parts_target']
                    total[d.part.id][COST] += float(part_ratio
                                                    * cost['overhead'])
            CACHE.set(CACHE_ID, total, timeout=None)

        cost = 0
        qty = 0
        for p in self.equivalent_parts():
            if p.id in total:
                cost += total[p.id][COST]
                qty += total[p.id][QTY]
        if qty <= 0:
            return self.target_price
        return part_cost(cost, qty)

    def best_vendor_part(self, preferred_vendors=None):
        """Find the best (lowest-cost and orderable) part from the specified
        list of preferred vendors (or any vendor if preferred_vendors
        is None).  If the part cannot be ordered from any of the
        preferred vendors, return the vendor part from the
        non-preferred vendor with the lowest cost.

        """
        best_vp_any = None
        best_vp_preferred = None
        for p in self.equivalent_parts():
            if not p.is_orderable():
                continue
            vps = Vendor_Part.objects.filter(part_id=p.id)
            for vp in vps:
                if not vp.is_orderable():
                    continue
                if best_vp_any is None or vp.price < best_vp_any.price:
                    best_vp_any = vp
                if preferred_vendors is None or \
                   vp.vendor not in preferred_vendors:
                    continue
                if best_vp_preferred is None \
                   or vp.price < best_vp_preferred.price:
                    best_vp_preferred = vp
        if best_vp_preferred:
            return best_vp_preferred
        return best_vp_any

    def strstatus(self):
        return strchoice(Enum.STATUS_CHOICES, self.status)

    def choice_label(self):
        return '%s: %s %s' % (self, self.mfg, self.mfg_pn)

class Vendor(models.Model):
    name = models.CharField(max_length=31, verbose_name='Vendor Name',
                            unique=True)
    search_url = models.CharField(max_length=127,
                                  verbose_name='Search URL Pattern',
                                  default='', blank=True,
                                  help_text='This pattern defines how to '
                                  'search for a particular part on the '
                                  'vendor\'s website.  %(vendor_pn)s gets '
                                  'replaced by the vendor\'s part-number, '
                                  '%(mfg)s by the manufacturer\'s name, and '
                                  '%(mfg_pn)s by the manufacturer\'s '
                                  'part-number.')
    def __str__(self):
        return '%s' % self.name

    def html_link(self):
        url = reverse('epic:vendor_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.name)

    def is_vendor(self):
        return True

    class Meta:
        ordering = ['name']
        indexes = [
            models.Index(fields=['name'])
        ]

class Vendor_Part(models.Model):
    part = models.ForeignKey(Part, verbose_name='Part #',
                             on_delete=models.CASCADE)
    vendor = models.ForeignKey(Vendor, on_delete=models.CASCADE)
    vendor_pn = models.CharField(max_length=31,
                                 verbose_name='Vendor\'s Part #')
    price = models.DecimalField(max_digits=10, decimal_places=6,
                                validators=PART_PRICE_VALIDATORS)
    status = models.IntegerField(choices=Enum.STATUS_CHOICES,
                                 default=Enum.STATUS_ACTIVE, blank=False,
                                 verbose_name='Life-time Status',
                                 help_text='The life-time status of this '
                                 'vendor part.')

    class Meta:
        unique_together = (
            ('vendor', 'part'),
            ('vendor', 'vendor_pn')
        )

    def __str__(self):
        return 'Vendor_Part %s: %s %s' % (format_part_number(self.part.id),
                                          self.vendor.name, self.vendor_pn)

    def html_link(self):
        if self.vendor.search_url == '':
            return escape(self.vendor_pn)
        url = self.vendor.search_url % \
              {
                  'mfg'		: escape(self.part.mfg),
                  'mfg_pn'	: escape(self.part.mfg_pn),
                  'vendor_pn'	: escape(self.vendor_pn)
              }
        return '<a href="%s" target="part_lookup">%s</a>' % \
            (url, self.vendor_pn)

    def strstatus(self):
        return strchoice(Enum.STATUS_CHOICES, self.status)

    def is_orderable(self):
        return self.status in Enum.STATUS_ORDERABLE

    @staticmethod
    def get(part_id, vendor_id):
        result = Vendor_Part.objects \
                            .filter(part_id=part_id) \
                            .filter(vendor_id=vendor_id)
        if len(result) != 1:
            return None
        return result[0]

class Warehouse(models.Model):
    name = models.CharField(max_length=31, unique=True,
                            help_text='The name of the warehouse.')
    address = models.TextField(blank=True, default='', help_text='The '
                               'shipping address for the warehouse.')

    def __str__(self):
        return '%s' % self.name

    def html_link(self):
        url = reverse('epic:warehouse_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.name)

    def inventories(self):
        return Inventory.objects.filter(warehouse_id=self.id)

    @staticmethod
    def by_name(name):
        return Warehouse.objects.get(name=name)

    class Meta:
        ordering = ['name']

class Delta(models.Model):
    """Deltas are used to track all quantity adjustments to a part at a
    given warehouse.  The adjustment may be relative to existing
    quantity or absolute.
    """
    part = models.ForeignKey(Part, verbose_name='Part #',
                             help_text='The part whose quantity gets '
                             'adjusted.', on_delete=models.CASCADE)
    is_absolute = models.BooleanField(verbose_name='Absolute Adjustment',
                                      default=False,
                                      help_text='If set, the adjustment is '
                                      'absolute otherwise it is relative.')
    adj = models.IntegerField(verbose_name='Adjustment Count',
                              help_text='The amount by which the part '
                              'quantity should be adjusted by.')
    txtn = models.ForeignKey('Transaction', verbose_name='Transaction #',
                             on_delete=models.CASCADE)
    warehouse = models.ForeignKey(Warehouse, on_delete=models.CASCADE)

    class Meta:
        indexes = [
            models.Index(fields=['txtn']),
            models.Index(fields=['warehouse'])
        ]

    def line_item(self):
        """For order or shipments items, return associated line-item."""
        line_items = Line_Item.objects.filter(txtn_id=self.txtn.id) \
                                      .filter(part_id=self.part_id)
        if len(line_items) != 1:
            return None
        return line_items[0]

    def line_cost(self):
        """For order or shipments items, return line-cost for this delta."""
        line_item = self.line_item()
        if line_item is None:
            return None
        return Decimal('%.2f' % (self.adj *
                                 float(line_item.line_cost) / line_item.qty))

    def __str__(self):
        return 'Delta %s' % self.id

class Assembly_Item(models.Model):
    assy = models.ForeignKey(Part, related_name='assembly_item_part',
                             verbose_name='Assembly Part #',
                             help_text='The part number of the assembly this '
                             'item belongs to.', on_delete=models.CASCADE)
    comp = models.ForeignKey(Part, related_name='assembly_item_comp',
                             verbose_name='Component Part #',
                             help_text='The part number of the component of '
                             'this item.', on_delete=models.CASCADE)
    qty = models.PositiveIntegerField(help_text='The quantity of this '
                                      'component required by the assembly.')
    refdes = models.TextField(blank=True, default='',
                              verbose_name='Refdes List',
                              help_text='A list of comma-separated reference '
                              'designators e.g., "R101,R304". '
                              'The length of this list should match Quantity.')

    class Meta:
        unique_together = ('assy', 'comp')
        index_together = [
            ['assy', 'comp']
        ]

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        if self.assy is not None:
            qs = Stock_Cache_Entry \
                 .objects \
                 .filter(stock_cache_item__part_id=self.assy.id)
            if qs.exists():
                Stock_Cache_Entry.invalidate(qs.order_by('ts')[0].ts)

    def qty_with_overage(self, factor):
        q = math.ceil(abs(factor) * self.qty * (1 + self.comp.overage / 100))
        return int(math.copysign(q, factor))

    def __str__(self):
        return 'Assembly_Item %s.%s' % (format_part_number(self.assy_id),
                                        format_part_number(self.comp_id))

    def clean(self):
        refdes_list = re.findall(r'([^, ]+)', self.refdes)
        self.refdes = ','.join(refdes_list)

class Line_Item(models.Model):
    txtn = models.ForeignKey('Transaction', verbose_name='Transaction #',
                             on_delete=models.CASCADE)
    part = models.ForeignKey(Part, verbose_name='Part #',
                             on_delete=models.CASCADE)
    qty = models.PositiveIntegerField(help_text='The quantity of this part.')
    line_cost = models.DecimalField(max_digits=9, decimal_places=2,
                                    verbose_name='Amount',
                                    help_text='The total cost of this '
                                    'line-item.  Part-cost is line-item cost '
                                    'divided by quantity.',
                                    validators=[MinValueValidator(
                                        Decimal('0.00'))])
    index = models.PositiveIntegerField(help_text='Sequential and unchanging '
                                        'index of this line-item.')
    class Meta:
        ordering = ['index']
        unique_together = (
            ('txtn', 'part'),
            ('txtn', 'index')
        )
        index_together = [
            ['txtn', 'part'],
            ['txtn']
        ]

    def vendor_part(self):
        result = Vendor_Part.objects \
                            .filter(part_id=self.part_id) \
                            .filter(vendor_id=self.txtn.order.vendor.id)
        if len(result) != 1:
            return None
        return result[0]

    def part_cost(self):
        return part_cost(self.line_cost, self.qty)

    def order_line_item(self):
        """For shipment line-items, return associated order line-item."""
        line_items = Line_Item.objects \
                              .filter(txtn_id=self.txtn.shipment.ordr.id) \
                              .filter(part_id=self.part_id)
        if len(line_items) != 1:
            return None
        return line_items[0]

    def qty_remaining_to_ship(self):
        if not hasattr(self.txtn, 'order'):
            return None

        shipment_ids = Shipment.objects.filter(ordr_id=self.txtn_id) \
                                       .values_list('transaction_ptr_id',
                                                    flat=True)
        already_shipped = sum(Line_Item.objects \
                              .filter(txtn_id__in=shipment_ids) \
                              .filter(part_id=self.part_id) \
                              .values_list('qty', flat=True))
        return self.qty - already_shipped

class Transaction(models.Model):
    """Orders, shipments, and inventories are all transactions."""
    ts = models.DateTimeField(verbose_name='Creation Time',
                              help_text='Date and time when this transaction '
                              'was created.')
    warehouse = models.ForeignKey(Warehouse,
                                  verbose_name='Warehouse',
                                  help_text='The (destination) warehouse '
                                  'this transaction applies to.',
                                  on_delete=models.CASCADE)
    notes = models.TextField(blank=True,
                             verbose_name='Notes',
                             help_text='Comments and notes for this '
                             'transaction')

    def __str__(self):
        return 'Transaction %u' % (self.id)

    def save(self, *args, **kwargs):
        if self.id is not None:
            qs = Transaction.objects.filter(pk=self.id)
            if qs.exists():
                old = qs[0]

                if hasattr(self, 'shipment'):
                    # delete all deltas for this transaction, if any;
                    # the new deltas will be created during the
                    # update_deltas() call
                    Delta.objects.filter(txtn_id=old.id).delete()

                # invalidate stock-cache entries not older than the minimum
                # old or new ts of this transaction:
                ts = min(self.ts, old.ts)
                Stock_Cache_Entry.invalidate(ts)
        super().save(*args, **kwargs)

    def first_available_index(self):
        max_index = 0
        for item in Line_Item.objects.filter(txtn_id=self.id):
            if item.index > max_index:
                max_index = item.index
        return max_index + 1

    def update_order_status(self):
        if hasattr(self, 'order'):
            # transaction is an order...
            order = self.order
            # we ought to be able to check if the transaction is a
            # shipment by checking for hasattr(self, 'shipment').
            # However, that does not work if the shipment has been
            # deleted AND self.shipment has not been accessed before.
            # Fortunately, 'ordr' exists in all cases and is also
            # unique to a shipment:
        elif hasattr(self, 'ordr'):
            # transaction is a shipment
            order = self.ordr
            if order is None:
                return	# inter-warehouse shipment
        else:
            return

        status = Enum.TXTN_STATUS_CLOSED
        for line_item in Line_Item.objects.filter(txtn_id=order.id):
            if line_item.qty_remaining_to_ship() > 0:
                status = Enum.TXTN_STATUS_OPEN
                break

        if status != order.status:
            order.status = status
            order.save()

    def update_deltas(self):
        if hasattr(self, 'order') or hasattr(self, 'inventory'):
            # orders never have deltas and inventories directly edit the deltas
            return

        if hasattr(self, 'shipment'):
            # For shipments that have not been deleted yet, create a
            # delta for each line-item in the shipment:
            ship = self.shipment
            src_warehouse = ship.src_warehouse()

            for line_item in Line_Item.objects.filter(txtn_id=ship.id):
                d = Delta(part_id=line_item.part_id, is_absolute=False,
                          adj=line_item.qty, txtn_id=self.id,
                          warehouse_id=ship.warehouse_id)
                d.save()

                assy_items = d.part.assembly_items()
                if ship.ordr and assy_items.exists():
                    # assemblies ordered from an assembly-house are created
                    # from parts at the assembly-house:
                    for assy_item in assy_items:
                        qty = -assy_item.qty_with_overage(line_item.qty)
                        part = assy_item.comp.best_part()
                        d = Delta(part_id=part.id, is_absolute=False,
                                  adj=qty, txtn_id=self.id,
                                  warehouse_id=src_warehouse.id)
                        d.save()
                elif ship.from_warehouse is not None:
                    d = Delta(part_id=line_item.part_id, is_absolute=False,
                              adj=-line_item.qty, txtn_id=self.id,
                              warehouse_id=src_warehouse.id)
                    d.save()

    def finalize(self):
        self.update_order_status()
        self.update_deltas()

    def description(self):
        if hasattr(self, 'order'):
            ttl = self.order.days_until_past_due()
            extra = ''
            if ttl and ttl <= 0:
                extra = ' (expected %d days ago)' % -ttl
            desc = '%s PO %s shipping to %s%s.' % \
                (self.order.vendor.html_link(), self.order.html_link(),
                 self.warehouse.html_link(), extra)
        elif hasattr(self, 'shipment'):
            ship = self.shipment
            if ship.ordr_id:
                desc = 'Shipment %s to %s for %s PO %s.' % \
                       (ship.html_link(), ship.warehouse.html_link(),
                        ship.ordr.vendor.html_link(), ship.ordr.html_link())
            else:
                desc = 'Shipment %s from %s to %s.' % \
                       (ship.html_link(), ship.from_warehouse.html_link(),
                        ship.warehouse.html_link())
        elif hasattr(self, 'inventory'):
            desc = 'Inventory %s at %s.' % (self.inventory.html_link(),
                                            self.warehouse.html_link())
        else:
            desc = 'Unknown transaction %d.' % self.id
        return desc

    def html_link(self):
        if hasattr(self, 'order'):
            return 'Order&nbsp;%s' % self.order.html_link()
        if hasattr(self, 'shipment'):
            return 'Shipment&nbsp;%s' % self.shipment.html_link()
        if hasattr(self, 'inventory'):
            return 'Inventory&nbsp;%s' % self.inventory.html_link()
        return 'Transaction&nbsp;%d' % (self.id)

class Order(Transaction):
    expected_arrival_date = models.DateField(verbose_name='Expected Arrival '
                                             'Date', help_text='Date when '
                                             'the order is expected to '
                                             'arrive.')
    status = models.IntegerField(choices=Enum.TXTN_STATUS_CHOICES,
                                 default=Enum.TXTN_STATUS_OPEN, blank=False,
                                 verbose_name='Order Status')
    vendor = models.ForeignKey(Vendor, help_text='The name of the '
                               'vendor (distributor) where the order '
                               'was placed.', on_delete=models.CASCADE)

    def __str__(self):
        return '%u' % (self.id)

    def total_cost(self):
        return sum(Line_Item.objects
                   .filter(txtn_id=self.id) \
                   .values_list('line_cost', flat=True))

    def is_open(self):
        return self.status == Enum.TXTN_STATUS_OPEN

    def days_until_past_due(self):
        if self.status != Enum.TXTN_STATUS_OPEN:
            return None
        return (self.expected_arrival_date - date.today()).days

    def assembly_line_items(self):
        """Returns the list of line-items that are assemblies."""
        assemblies = Assembly_Item.objects.values_list('assy_id', flat=True) \
                                          .distinct()
        return Line_Item.objects.filter(txtn_id=self.id) \
                                .filter(part_id__in=assemblies)

    def html_link(self):
        url = reverse('epic:order_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)

    def choice_label(self):
        return '%s: %s $%s' % (self.id, self.vendor.name,
                               intcomma(self.total_cost()))

class Shipment(Transaction):
    tracking = models.CharField(max_length=127, default='', blank=True,
                                verbose_name='Tracking #s',
                                help_text='Comma-separated list of tracking '
                                'numbers.')
    # order is a reserved word in django...
    ordr = models.ForeignKey(Order, verbose_name='Order #',
                             null=True, blank=True,
                             help_text='For an order shipment, the '
                             'order that resulted in this shipment.',
                             on_delete=models.CASCADE)
    from_warehouse = models.ForeignKey(Warehouse, null=True, blank=True,
                                       help_text='For an inter-warehouse '
                                       'shipment, the warehouse '
                                       'the shipment originates from.',
                                       on_delete=models.CASCADE)
    cost_freight = models.DecimalField(max_digits=9, decimal_places=2,
                                       verbose_name='Freight Cost',
                                       validators=[MinValueValidator(
                                           Decimal('0.00'))])
    cost_other = models.DecimalField(max_digits=9, decimal_places=2,
                                     verbose_name='Other Costs',
                                     help_text='Other costs assessed by the '
                                     'shipper, such as handling costs.',
                                     validators=[MinValueValidator(
                                         Decimal('0.00'))])
    cost_discount = models.DecimalField(max_digits=9, decimal_places=2,
                                        verbose_name='Discount Given',
                                        help_text='Discounts given by the '
                                        'shipper, such as early payment '
                                        'discount.',
                                        validators=[
                                            MinValueValidator(
                                                Decimal('0.00'))])

    @classmethod
    def flush_caches(cls, sender, **kwargs):
        """See signals.flush_caches()."""
        # pylint: disable=unused-argument
        # cache of shipment costs, indexed by transaction id
        CACHE.clear()

    def __str__(self):
        return 'Shipment %u' % self.id

    def clean(self):
        tracking_list = re.findall(r'([^, ]+)', self.tracking)
        self.tracking = ','.join(tracking_list)

    def html_link(self):
        url = reverse('epic:ship_detail', kwargs={'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)

    def html_tracking_links(self):
        trk_list = self.tracking.split(',')
        url_pattern = 'https://www.packagetrackr.com/track/%s'
        result = []
        for trk in trk_list:
            trk = trk.strip()
            url = url_pattern % trk
            result.append('<a target="trk_lookup" href="%s">%s</a>' \
                           % (url, trk))
        return html_list_or_none(result)

    def src_warehouse(self):
        if self.ordr is not None:
            try:
                return Warehouse.by_name(self.ordr.vendor)
            except Exception:
                return None
        else:
            return self.from_warehouse

    def overhead_cost(self):
        """Return the cost of overheads of this shipment."""
        return self.cost_freight + self.cost_other - self.cost_discount

    def cost_details(self):
        costs = CACHE.get('costs-' + str(self.id))
        if costs is None:
            parts_cost = 0
            parts_target_cost = 0
            part_detail = {}
            for line_item in Line_Item.objects.filter(txtn_id=self.id):
                if part_detail is not None:
                    part_detail[line_item.part_id] = {
                        'cost': line_item.line_cost,
                        'qty': line_item.qty
                    }
                if self.from_warehouse is None:
                    parts_cost += line_item.line_cost
                else:
                    # for inter-warehouse shipments, each line_item.line_cost
                    # is zero but we need an approximation of part cost so
                    # we can allocate the cost of the shipment properly
                    parts_target_cost += line_item.qty \
					 * line_item.part.target_price
            costs = {
                'overhead': self.overhead_cost(),
                'parts': parts_cost,
                'parts_target': parts_target_cost,
                'part_detail': part_detail
            }
            CACHE.set('costs-' + str(self.id), costs, timeout=None)
        return costs

    def parts_cost(self):
        """Return the total cost of all parts in this shipment."""
        return self.cost_details()['parts']

    def total_cost(self):
        return self.overhead_cost() + self.parts_cost()

class Inventory(Transaction):
    def __str__(self):
        return 'Inventory %u' % self.id

    def html_link(self):
        url = reverse('epic:warehouse_inventory_detail',
                      kwargs={'warehouse': self.warehouse.id, 'pk': self.id})
        return '<a href="%s">%s</a>' % (url, self.id)

class Part_Count:
    # pylint: disable=too-few-public-methods
    def __init__(self, qty=0):
        self.qty = qty
        self.is_final = False

class Part_Inventory:
    def __init__(self, ts=None):
        self.warehouse = {}
        self.ts = ts

    def __str__(self):
        lines = []
        lines.append('Inventory as of %s:' % self.ts)
        for w in self.warehouse:
            s = ''
            for p in self.warehouse[w]:
                if self.warehouse[w][p].qty != 0:
                    s += (' %s*%d' % (format_part_number(p),
                                      self.warehouse[w][p].qty))
            lines.append('  Warehouse %s:%s' % (w, s))
        return '\n'.join(lines)

    def get_entry(self, warehouse_id, part_id):
        if warehouse_id in self.warehouse:
            entry = self.warehouse[warehouse_id]
        else:
            self.warehouse[warehouse_id] = entry = {}
        if part_id not in entry:
            entry[part_id] = Part_Count()
        return entry[part_id]

    def apply_assembly_item(self, warehouse_id, item, factor):
        # Note: Even though this component might be an assembly itself,
        #  there is no need to recurse down because parts do not
        #  spontaneously transform into assemblies.  That is, even if
        #  we had the pre-requisite sub-components, there still has to
        #  be an assembly order which is where we'd do the proper
        #  sub-component qty adjustments.
        part_id = item.comp.best_part().id
        qty = item.qty_with_overage(factor)
        entry = self.get_entry(warehouse_id, part_id)
        entry.qty += qty

    def apply_order_line_item(self, item):
        part_id = item.part_id
        order = item.txtn.order
        qty = item.qty_remaining_to_ship()
        if qty <= 0:
            return
        entry = self.get_entry(order.warehouse_id, part_id)
        entry.qty += qty

        assy_items = item.part.assembly_items()
        if assy_items.exists() > 0:
            vendor_id = Warehouse.by_name(order.vendor).id
            for assy_item in item.part.assembly_items():
                self.apply_assembly_item(vendor_id, assy_item, -qty)

    def apply_delta(self, delta, reverse_chronological_order=False):
        '''Adjust inventory according to DELTA.  If applying deltas in reverse
        chronological order (newest first), you must set
        REVERSE_CHRONOLOGICAL_ORDER to True and stop processing deltas
        after as the first delta with "is_absolute==True" has been
        processed.

        '''
        entry = self.get_entry(delta.warehouse_id, delta.part_id)
        if entry.is_final:
            return

        if delta.is_absolute:
            if reverse_chronological_order:
                entry.qty += delta.adj
                entry.is_final = True
            else:
                entry.qty = delta.adj
        elif not entry.is_final:
            entry.qty += delta.adj

    def get_qty(self, warehouse, part):
        if warehouse not in self.warehouse:
            return 0
        if part not in self.warehouse[warehouse]:
            return 0
        return self.warehouse[warehouse][part].qty

    def adj_qty(self, warehouse, part, adj):
        entry = self.get_entry(warehouse, part)
        entry.qty += adj

class Stock_Cache_Entry(models.Model):
    '''A stock cache entry lists all part counts of all warehouses as of a
    given point in time.  In contrast to inventories, such entrys are
    automatically created as needed and deleted whenever a delta
    changes that could affect the entry.

    '''
    ts = models.DateTimeField(verbose_name='Creation Time',
                              help_text='Date and time when this transaction '
                              'was created.')
    class Meta:
        ordering = ['-ts']
        indexes = [
            models.Index(fields=['ts']),
        ]

    @classmethod
    def invalidate(cls, ts=None):
        '''Invalidate all stock cache entries that are not older than the
        specified timestamp.  All entries are invalidated if the
        timestamp is None.

        '''
        Stock_Cache_Entry.objects.filter(ts__gte=ts).delete()

    def __str__(self):
        return 'Stock_Cache_Entry %u' % self.id

    def inventory(self):
        inv = Part_Inventory(ts=self.ts)
        qs = Stock_Cache_Item.objects.filter(stock_cache_entry_id=self.id)
        for item in qs:
            entry = inv.get_entry(item.warehouse.id, item.part.id)
            entry.qty = item.qty
        return inv

    def save(self, *args, **kwargs):
        inventory = None
        if 'inventory' in kwargs:
            inventory = kwargs['inventory']
            kwargs.pop('inventory')
            self.ts = inventory.ts

        super().save(*args, **kwargs)

        # Save part counts in Stock_Cache_Item:
        if inventory is not None:
            for warehouse in inventory.warehouse:
                for part_id in inventory.warehouse[warehouse]:
                    qty = inventory.warehouse[warehouse][part_id].qty
                    if qty != 0:
                        item = Stock_Cache_Item(stock_cache_entry=self,
                                                warehouse_id=warehouse,
                                                part_id=part_id,
                                                qty=qty)
                        item.save()

class Stock_Cache_Item(models.Model):
    stock_cache_entry = models.ForeignKey(Stock_Cache_Entry,
                                          on_delete=models.CASCADE)
    warehouse = models.ForeignKey(Warehouse,
                                  verbose_name='Warehouse',
                                  help_text='The warehouse stocking '
                                  'this part.',
                                  on_delete=models.CASCADE)
    part = models.ForeignKey(Part, verbose_name='Part #',
                             on_delete=models.CASCADE)
    qty = models.IntegerField(help_text='The quantity of this part.')

    class Meta:
        indexes = [
            models.Index(fields=['stock_cache_entry']),
        ]

    def __str__(self):
        return 'Stock_Cache_Item[%s]: %s*%s' % \
            (self.warehouse.name, format_part_number(self.part.id), self.qty)



#!/usr/bin/env python
#
#   Copyright (c) 2014-2021 eGauge Systems LLC
#	1644 Conestoga St, Suite 2
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''This is a KiCad Eeschema plugin that converts a KiCad XML to:

	1) a bill-of-materials (BOM), output as CSV text file
	   (comma-separated values)
	2) an EPIC part (assembly) that consists of the BOM components

The EPIC assembly created by this tool will have its manufacturer set
to django.conf.EPIC_MANUFACTURER and the manufacturer's part-number
will be set to 'bom:' with the name of the schematic and flavor-name
(if any) appended.  If an assembly with that part-number already
exists, it will be updated unless the BOM was edited manually inside
EPIC.  In the latter case, you can force overwriting the manually
edited BOM pass passing this script the --force-overwrite command-line
option.

This tool also updates the Eeschema schematic files with the footprint
info in the EPIC database.  The original schematic files are backed up
in files with a tilde (~) appended to the original filename.  If the
footprint in a schematic file changes, the tool will warn you about
that and remind you to re-generate the net list file and potentially
update the PCB.

This tool extracts part number an installation information from the
XML file using the following component field names:

	Field Name:	Purpose:
	------------	---------------------------------------------
	EP		Specifies the EPIC part number to use for the
			component.

	Installation	If the value of this field is "DNP" (Do Not Place),
			then the component is omitted from the BOM.
			In the CSV output file, do-not-place components
			are listed separately at the end of the file.

A single Eeschema file may define multiple flavors (versions) of a
BOM.  This is done by appending the flavor name to the above field
names, separated by a dash (-).  For example, to define a flavor
called `lite', you'd use field names `EP-lite', `Installation-lite',
or `Value-lite' when a particular component needs to have a different
EPIC part number, installation method, or value for that flavor,
respectively.  Use command-line option --flavor to specify which
flavor to generate the a BOM for.

'''

# We have to import django after setting DJANGO_SETTINGS_MODULE and
# pylint doesn't like that:
# pylint: disable=wrong-import-position

from __future__ import print_function

from types import SimpleNamespace

import argparse
from argparse import RawDescriptionHelpFormatter as RawDescHelpFormatter
import io
import os
from os import path
import re
import sys
import tempfile
import xml.etree.ElementTree as ET

import sexpdata

# Please replace 'epic-sample' with the name of your Django project:
os.environ.setdefault('DJANGO_SETTINGS_MODULE',
                      'epic-sample.epic-sample.settings')

import django
django.setup()

from django.conf import settings as cfg
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction

from epic.models import Part, Assembly_Item, Vendor
from epic.base import Enum, format_part_number, strchoice

MY_TOOL_NAME = 'kicad-to-epic-bom'

KICAD_FOOTPRINT = sexpdata.Symbol('footprint')
KICAD_PATH = sexpdata.Symbol('path')
KICAD_PROP = sexpdata.Symbol('property')
KICAD_PROP_EP = 'EP'				# eGauge part number
KICAD_PROP_FOOTPRINT = 'Footprint'
KICAD_PROP_REFDES = 'Reference'
KICAD_SYMBOL = sexpdata.Symbol('symbol')
KICAD_SYMBOL_INSTANCES = sexpdata.Symbol('symbol_instances')
KICAD_UUID = sexpdata.Symbol('uuid')

manufacturer = cfg.EPIC_MANUFACTURER

class Component:
    def __init__(self, part_id=None, refdes=None, value=None, footprint=None,
                 mfg=None, mfg_pn=None, part=None):
        self.part_id = part_id
        self.part = part
        self.refdes = refdes
        self.value = value
        self.footprint = footprint
        self.mfg = mfg
        self.mfg_pn = mfg_pn

    def __str__(self):
        res = 'Component('
        sep = ''
        for f in dir(self):
            if f[0] == '_':
                continue
            val = self.__dict__[f]
            if val is not None:
                res += '%s%s=%s' % (sep, f, val)
                sep = ','
        res += ')'
        return res

def quote(string):
    string = string.replace('"', '\"')
    # Replace comma by colon.  DigiKey doesn't
    # handle commas correctly even if they're inside double quotes...
    string = string.replace(',', ':')
    return '"' + string + '"'

@transaction.atomic
def save_assembly(assembly_part, bom):
    created = assembly_part.id is None

    assembly_part.last_bom_mod_type = Enum.LAST_MOD_TYPE_TOOL
    assembly_part.last_bom_mod_name = MY_TOOL_NAME
    assembly_part.save()

    Assembly_Item.objects.filter(assy_id=assembly_part.id).delete()

    for components in bom:
        comp = components[0]
        refdes = ','.join(sorted([c.refdes for c in components]))
        item = Assembly_Item(assy_id=assembly_part.id,
                             comp_id=comp.part_id,
                             qty=len(components),
                             refdes=refdes)
        item.save()
    print('%s: info: %s part %s (%s %s) with %d components' \
          % (MY_TOOL_NAME, 'created' if created else 'updated',
             format_part_number(assembly_part.id),
             assembly_part.mfg, assembly_part.mfg_pn, len(bom)),
          file=sys.stderr)

def legacy_update_comp(comp):
    '''Check the footprint of a component in legacy format and update it
    if necessary.  COMP is the component to update in the form of a
    list of text lines.  Returns True if the component was updated,
    False otherwise.

    '''
    footprint_field_index = part_number = None
    has_changed = False
    refdes = ''
    for index, line in enumerate(comp):
        m = re.match(r'F\s+(\d+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+'
                     r'\s+\S+(\s+(\S+))?$', line)
        if m:
            if m.group(1) == '0':
                if m.lastindex >= 2:
                    refdes = m.group(2)[1:-1] + ' '
            elif m.group(1) == '2':
                footprint_field_index = index
            elif m.lastindex >= 3 and m.group(4) == '"EP"':
                try:
                    part_str = m.group(2)[1:-1]
                    part_number = int(part_str)
                except ValueError:
                    print('%s: warn: invalid EPIC part number "%s"' %
                          (MY_TOOL_NAME, part_str), file=sys.stderr)
    if footprint_field_index is not None and part_number is not None:
        try:
            part = Part.objects.get(pk=part_number)
        except ObjectDoesNotExist:
            print('%s: error: Part %s does not exist' %
                  (MY_TOOL_NAME, format_part_number(part_number)),
                  file=sys.stderr)
            return False
        if part.footprint is None or part.footprint == '':
            if part.mounting != Enum.MOUNTING_CHASSIS \
               and part.mounting != Enum.MOUNTING_FREE:
                print('%s: warn: EPIC part %s has no footprint defined' %
                      (MY_TOOL_NAME, format_part_number(part.id)),
                      file=sys.stderr)
        else:
            old = comp[footprint_field_index]
            m = re.match(r'F\s+2\s+"([^"]*)"+\s+(.*)', old)
            if m:
                old_footprint = m.group(1)
                rest = m.group(2)
                new_footprint = part.footprint
                new = 'F 2 \"%s\" %s\n' % (new_footprint, rest)
                comp[footprint_field_index] = new
                if old_footprint != new_footprint:
                    if old_footprint == '':
                        msg = 'set'
                    else:
                        msg = 'changed from %s' % old_footprint
                    print('%s: warn: %sfootprint %s to %s.'
                          % (MY_TOOL_NAME, refdes, msg, new_footprint))
                    has_changed = True
                else:
                    has_changed = False
    return has_changed

def legacy_update_schematic(path, out):
    '''Update the component footprints in a schematic in legacy format
    (pre KiCad v6) as needed.  PATH is the filename path of the
    schematic, OUT is a temporary output file to write the updated
    schematic file to.  Returns True if the schematic file changed,
    False otherwise.

    '''
    has_changes = False
    curr_comp = []
    with io.open(path, 'r', encoding='utf-8') as schem_file:
        inside_comp = False
        for line in schem_file:
            if line.rstrip() == '$Comp':
                inside_comp = True
                curr_comp.append(line)
            elif line.rstrip() == '$EndComp':
                inside_comp = False
                curr_comp.append(line)
                has_changes |= legacy_update_comp(curr_comp)
                for comp_line in curr_comp:
                    out.write(comp_line)
                curr_comp = []
            elif inside_comp:
                curr_comp.append(line)
            else:
                out.write(line)
    return has_changes

def modern_update_symbol(symbol, updated_syms):
    '''Check the footprint of a schematic symbol and update it if
    necessary.  SYMBOL is the s-expression of the symbol to update.
    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
    have been updated.  The value of each entry is the new footprint
    for that symbol.  Returns True if the symbol was updated, False
    otherwise.

    '''
    has_changed = False
    part_number = footprint_sexp = None
    refdes = ''
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_UUID:
            uuid = sexp[1].value()
        elif sexp[0] == KICAD_PROP:
            prop_name = sexp[1]
            if prop_name == KICAD_PROP_EP:
                try:
                    part_number = int(sexp[2])
                except ValueError:
                    print('%s: warn: invalid EPIC part number "%s"' %
                          (MY_TOOL_NAME, sexp[2]), file=sys.stderr)
            elif prop_name == KICAD_PROP_FOOTPRINT:
                footprint_sexp = sexp
            elif prop_name == KICAD_PROP_REFDES:
                refdes = sexp[2] + ' '

    if part_number is not None:
        try:
            part = Part.objects.get(pk=part_number)
        except ObjectDoesNotExist:
            print('%s: error: Part %s does not exist' %
                  (MY_TOOL_NAME, format_part_number(part_number)),
                  file=sys.stderr)
            return False
        footprint = part.footprint
        if not footprint:
            footprint = ''
            if part.mounting not in [Enum.MOUNTING_CHASSIS, Enum.MOUNTING_FREE]:
                print('%s: warn: EPIC part %s has no footprint defined' %
                      (MY_TOOL_NAME, format_part_number(part.id)),
                      file=sys.stderr)

        if footprint_sexp is None:
            print('%s: error: Footprint property missing in symbol!'
                  % (MY_TOOL_NAME), file=sys.stderr)
            return False

        old_footprint = footprint_sexp[2]
        if old_footprint != footprint:
            footprint_sexp[2] = footprint
            if not old_footprint:
                msg = 'set'
            else:
                msg = 'changed from %s' % old_footprint
            print('%s: warn: %sfootprint %s to %s.'
                  % (MY_TOOL_NAME, refdes, msg, footprint))
            updated_syms[uuid] = footprint
            has_changed = True
    return has_changed

def update_symbol_instance(symbol, new_footprint):
    '''Update a symbol instance whose footprint has changed.  SYMBOL is
    the s-expression of the symbol to update.  NEW_FOOTPRINT is the
    name of footprint to set for the symbol.

    '''
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_FOOTPRINT:
            sexp[1] = new_footprint
            return
    symbol.append([KICAD_FOOTPRINT, new_footprint])

def update_symbol_instances(symbol, updated_syms):
    '''Check SYMBOL's UUID is in UPDATED_SYMS and, if so, update its
    footprint with the name given as the value of the entry in
    UPDATED_SYMS.  Returns True if the symbol was updated, False
    otherwise.

    '''
    changed = False
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_PATH:
            path = sexp[1].split('/')
            sym_path = path[-1]
            if sym_path in updated_syms:
                update_symbol_instance(sexp, updated_syms[sym_path])
                changed = True
    return changed

def modern_update_schematic(path, out, updated_syms):
    '''Update the schematic symbol footprints in a schematic in modern
    format (post KiCad v5) as needed. PATH is the filename path of the
    schematic, OUT is a temporary output file to write the updated
    schematic file to.  UPDATED_SYMS is a dictionary of all symbol
    UUIDs whose footprints have been updated.  The value of each entry
    is the new footprint for that symbol.  Returns True if the
    schematic file changed, False otherwise.

    '''
    changed = False
    with io.open(path, 'r', encoding='utf-8') as schem_file:
        schematic = sexpdata.load(schem_file)
        for sexp in schematic:
            if isinstance(sexp, list) and sexp[0] == KICAD_SYMBOL:
                if modern_update_symbol(sexp, updated_syms):
                    changed = True
    if changed:
        sexpdata.dump(schematic, out)
    return changed

def update_schematic(dirname, filename, updated_syms):
    '''Update schematic file FILENAME residing in directory DIRNAME with
    footprints from the EPIC database.  UPDATED_SYMS is a dictionary
    into which are entered all symbol UUIDs whose footprints have been
    updated.  The value of each entry is the new footprint for that
    symbol.  Returns True if the schematic was updated, False
    otherwise.

    '''
    schem_path = os.path.join(dirname, filename)
    with tempfile.NamedTemporaryFile(dir=dirname, mode='w',
                                     encoding='utf-8') as out:
        _, ext = os.path.splitext(schem_path)
        if ext == '.sch':
            has_changes = legacy_update_schematic(schem_path, out)
        else:
            has_changes = modern_update_schematic(schem_path, out,
                                                  updated_syms)
        if has_changes:
            os.rename(schem_path, schem_path + '~')
            os.link(os.path.join(dirname, out.name), schem_path)
            print('%s: info: %s updated.  Use kicad to update PCB.'
                  % (MY_TOOL_NAME, schem_path))
    return has_changes

def update_schematic_sym_list(dirname, filename, has_backup, updated_syms):
    '''Update the footprints in the symbol-instances list of the top-level
    schematic of a KiCad v6 or newer schematic file.  DIRNAME is the
    directory containing the schematic and FILENAME is the filename
    within that directory.  HAS_BACKUP is True if the schematic file
    already has been backed up (due to an earlier change).
    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
    have been updated.  The value of each entry is the new footprint
    for that symbol.

    '''
    schem_path = os.path.join(dirname, filename)
    with tempfile.NamedTemporaryFile(dir=dirname, mode='w',
                                     encoding='utf-8') as out, \
         io.open(schem_path, 'r', encoding='utf-8') as schem_file:
        schematic = sexpdata.load(schem_file)
        changed = False
        for sexp in schematic:
            if isinstance(sexp, list) \
               and sexp[0] == KICAD_SYMBOL_INSTANCES:
                if update_symbol_instances(sexp, updated_syms):
                    changed = True

        if changed:
            sexpdata.dump(schematic, out)
            if not has_backup:
                os.rename(schem_path, schem_path + '~')
            os.link(os.path.join(dirname, out.name), schem_path)
            if not has_backup:
                print('%s: info: %s updated.  Use kicad to update PCB.'
                      % (MY_TOOL_NAME, schem_path))

def bom_append(bom, component):
    if component.part_id not in bom:
        bom[component.part_id] = []
    bom[component.part_id].append(component)

def bom_sort(bom):
    bom_list = []
    for part_id in bom:
        bom_list.append(bom[part_id])
    bom_list.sort(key=lambda x: x[0].part_id)
    return bom_list

def output_csv(outfile, bom_list, with_vendor_pn=False, preferred_vendors=None):
    for components in bom_list:
        qty = len(components)
        refdes = ','.join(sorted([c.refdes for c in components]))
        c = components[0]

        substitutes = ''
        if c.part:
            sep = ''
            for sub in c.part.equivalent_parts():
                if sub == c.part:
                    continue
                substitutes += sep
                substitutes += '%s (%s %s)' % (format_part_number(sub.id),
                                               sub.mfg, sub.mfg_pn)
                sep = ', '


        row = '%s,%s,%s,%s,%s,%s,%s,%s' % \
              (qty, c.value, c.footprint, format_part_number(c.part_id),
               c.mfg, quote(c.mfg_pn), quote(substitutes), quote(refdes))

        if with_vendor_pn:
            vp = c.part.best_vendor_part(preferred_vendors)
            if vp is None:
                vendor = 'n/a'
                vendor_pn = 'n/a'
            else:
                vendor = vp.vendor
                vendor_pn = vp.vendor_pn
            row += ',%s,%s' % (vendor, vendor_pn)
        print(row, file=outfile)

def update(res, field, desired_flavor, detected_flavors):
    '''Process RES by matching FIELD.attrib['name'] against RES.PATTERN.
    If there is a match M, M.group(2) must evaluate to the name of the
    flavor of the field.  If this flavor matches the specified flavor,
    we have an exact match and RES.value is set to field.text, RES.name
    is set to field.attrib['name'].
    If the flavor of the field is empty and RES.value is None we have a
    default match and RES is updated like for an exact match.
    DETECTED_FLAVOR is a set of flavors found.
    '''
    m = res.pattern.match(field.attrib['name'])
    if m is None:
        return

    this_flavor = None
    if m.lastindex is not None:
        this_flavor = m.group(2)
        detected_flavors |= {this_flavor} # update set of all detected flavors

    if this_flavor == desired_flavor \
       or (this_flavor is None and res.value is None):
        res.name = field.attrib['name']
        res.value = field.text

def main():
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=RawDescHelpFormatter)
    parser.add_argument('input-file', nargs=1, help='Name of XML file '
                        'generated by KiCad\'s \"Generate BOM\" command.')
    parser.add_argument('output-file', nargs=1, help='Name of output file '
                        'that will contain the bill-of-materials (BOM) in '
                        'a comma-separated-values (CSV) file.')
    parser.add_argument('--flavor', nargs=1, default=None,
                        help='Select PCB flavor for which to generate BOM '
                        'for.')
    parser.add_argument('-f', '--force-overwrite', action='store_true',
                        help='Force overwriting of existing EPIC BOM '
                        '(assembly) part even if the part appears to '
                        'have been modified manually.')
    parser.add_argument('-V', '--with-vendor-pn', action='store_true',
                        help='Include Vendor and Vendor part-number in the '
                        'CSV output.')
    parser.add_argument('-p', '--prefer-vendor', nargs='+', default=[],
                        help='Prefer the specified vendor(s), even if those '
                        'do not have the lowest cost for a given part. '
                        'Non-preferred vendors are used only for parts that '
                        'could be ordered otherwise.')
    parser.add_argument('-n', '--no-update-footprints', action='store_true',
                        help='By default, this program will update the '
                        'schematic files with the footprint information '
                        'found in the parts database.  This option can be '
                        'specified to skip this step.')
    parser.add_argument('--no-db-save', action='store_true',
                        help='Skip creating (or updating) the EPIC database '
                        'with the new BOM. '
                        'This is intended mainly for testing.')
    args = parser.parse_args()

    flavor = None
    if args.flavor is not None:
        flavor = args.flavor[0]

    preferred_vendors = None
    for vendor_name in args.prefer_vendor:
        vendors = Vendor.objects.filter(name__istartswith=vendor_name)
        if not vendors:
            print('%s: error: vendor `%s\' cannot be found' %
                  (MY_TOOL_NAME, vendor_name), file=sys.stderr)
            sys.exit(1)
        elif len(vendors) > 1:
            print('%s: error: vendor name `%s\' not unique, matches: %s' %
                  (MY_TOOL_NAME, vendor_name,
                   ', '.join(['`%s\'' % v.name for v in vendors])),
                  file=sys.stderr)
            sys.exit(1)
        if preferred_vendors is None:
            preferred_vendors = []
        preferred_vendors.append(vendors[0])

    output_filename = args.__dict__['output-file'][0]
    input_filename = args.__dict__['input-file'][0]

    xml = ET.parse(input_filename)

    schematic_name = 'unknown'
    revision = ''
    design = xml.find('design')
    sources = []
    source_path = path.dirname(input_filename)
    if design is not None:
        source = design.find('source')
        if source is not None:
            schematic_name = path.splitext(path.basename(source.text))[0]
        sheet = design.find('sheet')
        for sheet in design.iter('sheet'):
            title_block = sheet.find('title_block')
            if title_block is not None:
                if revision == '':
                    rev = title_block.find('rev')
                    if rev and rev.text:
                        revision = '-rev' + rev.text.lstrip().rstrip()
                sources.append(title_block.find('source').text)

    if not args.no_update_footprints:
        updated_syms = {}
        has_backup = False
        for schematic in sources:
            if update_schematic(source_path, schematic, updated_syms) \
               and schematic == sources[0]:
                has_backup = True
        if sources and updated_syms:
            update_schematic_sym_list(source_path, sources[0], has_backup,
                                      updated_syms)

    assembly_name = 'bom:' + schematic_name + revision
    if flavor is not None:
        assembly_name += '-' + flavor
    qs = Part.objects.filter(mfg=manufacturer).filter(mfg_pn=assembly_name)
    if qs.exists():
        assembly_part = qs[0]
        if assembly_part.last_bom_mod_type != Enum.LAST_MOD_TYPE_TOOL \
           or assembly_part.last_bom_mod_name != MY_TOOL_NAME:
            print('%s: info: part %s %s was last modified by %s %s' %
                  (MY_TOOL_NAME, manufacturer, assembly_name,
                   strchoice(Enum.LAST_MOD_CHOICES,
                             assembly_part.last_bom_mod_type),
                   assembly_part.last_bom_mod_name), file=sys.stderr)
            if args.force_overwrite:
                print('%s: info: overwriting existing part due to '
                      '--force-overwrite'
                      % MY_TOOL_NAME, file=sys.stderr)
            else:
                print('%s: error: refusing to overwrite part %s; '
                      'use --force-overwrite if desired' \
                      % (MY_TOOL_NAME, format_part_number(assembly_part.id)),
                      file=sys.stderr)
                sys.exit(1)
    else:
        desc = 'BOM for %s' % schematic_name
        if flavor is not None:
            desc += (', ' + flavor + ' flavor')
        assembly_part = Part(descr=desc,
                             mfg=manufacturer, mfg_pn=assembly_name,
                             mounting=Enum.MOUNTING_CHASSIS,
                             target_price=1000,
                             overage=1, spq=1, lead_time=4,
                             status=Enum.STATUS_PREVIEW)

    ep_pat = re.compile(r'EP(-(.*))?')
    inst_pat = re.compile(r'Installation(-(.*))?')
    value_pat = re.compile(r'Value(-(.*))?')

    bom = {}    # regular bom
    dnp = {}    # do-not-place bom
    # part number & refdes used to refer to a particular part:
    src_part_dict = {}
    detected_flavors = set()
    for comp in xml.find('components'):
        refdes = comp.attrib.get('ref')

        part = SimpleNamespace(name=None, value=None, pattern=ep_pat)
        inst = SimpleNamespace(name=None, value=None, pattern=inst_pat)
        value = SimpleNamespace(name=None, value=None, pattern=value_pat)
        fields = comp.find('fields')
        if fields is not None:
            for field in fields:
                update(part, field, flavor, detected_flavors)
                update(inst, field, flavor, detected_flavors)
                update(value, field, flavor, detected_flavors)

        do_not_place = (inst.value == 'DNP')

        if value.value is None:
            value.value = comp.findtext('value', default='n/a')

        if part.value is None:
            if not do_not_place:
                print('%s: warn: %s skipped due to missing '
                      'EPIC Part Number ("EP" field)'
                      % (MY_TOOL_NAME, refdes), file=sys.stderr)
            continue

        if do_not_place:
            print('%s: info: %s marked as do-not-place ("%s=DNP")'
                  % (MY_TOOL_NAME, refdes, inst.name), file=sys.stderr)

        try:
            part_id = int(part.value)
        except ValueError:
            print('%s: warn: %s has invalid EPIC part number "%s"' %
                  (MY_TOOL_NAME, refdes, part.value), file=sys.stderr)
            continue

        try:
            src_part = Part.objects.get(pk=part_id)
        except ObjectDoesNotExist:
            print('%s: warn: part %s not found in database' %
                  (MY_TOOL_NAME, format_part_number(part_id)), file=sys.stderr)
            continue

        c = Component(part_id, part=src_part, refdes=refdes, value=value.value)
        best_part = src_part.best_part()

        # keep a list of part names used to refer to best part:
        if best_part.id not in src_part_dict:
            src_part_dict[best_part.id] = [src_part.id, c.refdes]
        elif src_part.id != src_part_dict[best_part.id][0]:
            print('%s: warn: %s uses %s instead '
                  'of equivalent %s used by %s' % \
                  (MY_TOOL_NAME, c.refdes,
                   format_part_number(src_part.id),
                   format_part_number(src_part_dict[best_part.id][0]),
                   src_part_dict[best_part.id][1]))

        if c.value != c.part.val:
            print('%s: warn: %s has value "%s" but part %s has value "%s"'
                  % (MY_TOOL_NAME, c.refdes, c.value,
                     format_part_number(part_id), c.part.val), file=sys.stderr)
        c.footprint = c.part.footprint
        c.mfg = c.part.mfg
        c.mfg_pn = c.part.mfg_pn

        if do_not_place:
            bom_append(dnp, c)
        else:
            bom_append(bom, c)

    if not detected_flavors:
        detected_flavors_str = 'no flavors'
    else:
        if len(detected_flavors) > 1:
            detected_flavors_str = 'flavors: '
        else:
            detected_flavors_str = 'flavor: '
        detected_flavors_str += ', '.join(sorted(detected_flavors))

    if flavor is not None and flavor not in detected_flavors:
        print('%s: error: flavor `%s\' not defined by this schematic; '
              'detected %s'
              % (MY_TOOL_NAME, flavor, detected_flavors_str),
              file=sys.stderr)
        sys.exit(1)
    elif detected_flavors:
        print('%s: info: this schematic defines %s'
              % (MY_TOOL_NAME, detected_flavors_str), file=sys.stderr)

    bom_list = bom_sort(bom)
    dnp_list = bom_sort(dnp)

    out = io.open(output_filename, 'w', encoding='utf-8')
    hdr = (u'Qty,Value,Footprint,%s PN,Mfg,Mfg PN,Approved Substitutes,Refdes' %
           manufacturer)
    if args.with_vendor_pn:
        hdr += u',Vendor,Vendor PN'
    print(hdr, file=out)

    output_csv(out, bom_list, args.with_vendor_pn, preferred_vendors)

    if dnp_list:
        print('\nDO NOT PLACE parts:', file=out)
        output_csv(out, dnp_list, args.with_vendor_pn, preferred_vendors)

    if not args.no_db_save:
        save_assembly(assembly_part, bom_list)

if __name__ == '__main__':
    main()



#
#   Copyright (c) 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=too-few-public-methods, no-member, too-many-ancestors
# pylint: disable=no-self-use
import django_filters.rest_framework

from rest_framework import permissions, serializers, viewsets

from epic.models import Assembly_Item, Part, Vendor, Vendor_Part
from epic.perms import VIEW, EDIT

#
# Django query-ops supported when filtering based on object fields.
# For example:
#
#	api/assembly_item/?refdes__icontains=102
#
# returns all assembly-items whose refdes string includes "102".
#
INT_OPS = ['exact', 'lte', 'gte']
STRING_OPS = ['exact', 'startswith', 'istartswith', 'contains', 'icontains']

def get_bulk_update_serializer(self, *args, **kwargs):
    '''This method enables bulk-updating of models by setting the
    serializer's MANY argument to True when a JSON list is sent to the
    server.

    '''
    if 'data' in kwargs:
        data = kwargs['data']
        if isinstance(data, list):
            kwargs['many'] = True
    return super(viewsets.ModelViewSet, self).get_serializer(*args, **kwargs)


def expand_range(r):
    '''Convert a number-range string R to the corresponding set of numbers
    and return that set.  The range string must have the syntax:

      RANGE = DECIMAL ['-' DECIMAL].
      DECIMAL = DIGIT { DIGIT }.
      DIGIT = '0'-'9'.

    An empty set is returned if the R has invalid syntax.

    '''
    limits = r.split('-')
    try:
        if len(limits) == 1:
            return set([int(limits[0])])
        if len(limits) == 2:
            return set(range(int(limits[0]), int(limits[1]) + 1))
    except (TypeError, ValueError):
        pass
    return set()

def range_list_filter(fieldname, queryset, request):
    '''Filter QUERYSET according to a query-parameter called FIELDNAME.
    The query-parameters are taken from REQUEST, which must be the
    HTTP request that triggered this call.

    The query-parameter value specifies a set of numbers to be
    selected as a numeric range-list with syntax:

      RANGE_LIST = RANGE { ',' RANGE }.

    See method expand_range() for the RANGE syntax.

    For example, "id=13,20-22,99" would return the set containing
    numbers 13, 20, 21, 22, and 99.

    The queryset is then filtered to objects whose FIELDNAME value is
    in the specified set of numbers.

    '''
    range_list = request.query_params.get(fieldname)
    if range_list is not None:
        selected = set()
        for r in range_list.split(','):
            selected |= expand_range(r)
        params = {}
        params[fieldname + '__in'] = selected
        queryset = queryset.filter(**params)
    return queryset

class IsAuthorized(permissions.BasePermission):
    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return request.user.has_perms((VIEW,))
        return request.user.has_perms((EDIT,))

class PartSerializer(serializers.ModelSerializer):
    best_part = serializers.SerializerMethodField()
    equivalents = serializers.SerializerMethodField()

    class Meta:
        model = Part
        fields = '__all__'

    def get_best_part(self, obj):
        return obj.best_part().id

    def get_equivalents(self, obj):
        return [p.id for p in obj.equivalent_parts()]

class VendorSerializer(serializers.ModelSerializer):
    class Meta:
        model = Vendor
        fields = '__all__'

class VendorPartSerializer(serializers.ModelSerializer):
    class Meta:
        model = Vendor_Part
        fields = '__all__'

class AssemblyItemPartSerializer(serializers.ModelSerializer):
    class Meta:
        model = Assembly_Item
        fields = '__all__'

class PartViewSet(viewsets.ModelViewSet):
    '''API endpoint for viewing and editing parts.'''
    serializer_class = PartSerializer
    permission_classes = [IsAuthorized]
    filter_backends = (django_filters.rest_framework.DjangoFilterBackend, )
    filterset_fields = {
        'val': INT_OPS,
        'mfg': STRING_OPS,
        'mfg_pn': STRING_OPS,
        'mounting': INT_OPS,
        'status': INT_OPS
    }
    pagination_class = None

    get_serializer = get_bulk_update_serializer

    def get_queryset(self):
        return range_list_filter('id', Part.objects.all(), self.request)

    def update(self, request, pk=None):
        # delete assembly items associated with this part, if any:
        Assembly_Item.objects.filter(assy_id=pk).delete()
        return super().update(request, pk)

class VendorViewSet(viewsets.ModelViewSet):
    '''API endpoint for viewing and editing vendors.'''
    serializer_class = VendorSerializer
    permission_classes = [IsAuthorized]
    filter_backends = [django_filters.rest_framework.DjangoFilterBackend]
    filterset_fields = {
        'name': STRING_OPS
    }
    pagination_class = None

    get_serializer = get_bulk_update_serializer

    def get_queryset(self):
        return range_list_filter('id', Vendor.objects.all(), self.request)

class VendorPartViewSet(viewsets.ModelViewSet):
    '''API endpoint for viewing and editing vendor parts.'''
    serializer_class = VendorPartSerializer
    permission_classes = [IsAuthorized]
    filter_backends = [django_filters.rest_framework.DjangoFilterBackend]
    filterset_fields = {
        'part': INT_OPS,
        'vendor': INT_OPS,
        'vendor_pn': STRING_OPS,
        'status': INT_OPS
    }
    pagination_class = None

    get_serializer = get_bulk_update_serializer

    def get_queryset(self):
        qs = range_list_filter('id', Vendor_Part.objects.all(), self.request)
        return range_list_filter('part_id', qs, self.request)

class AssemblyItemViewSet(viewsets.ModelViewSet):
    '''API endpoint for viewing and editing assembly items.'''
    serializer_class = AssemblyItemPartSerializer
    permission_classes = [IsAuthorized]
    filter_backends = [django_filters.rest_framework.DjangoFilterBackend]
    filterset_fields = {
        'assy': INT_OPS,
        'comp': INT_OPS,
        'qty': INT_OPS,
        'refdes': STRING_OPS
    }
    pagination_class = None

    get_serializer = get_bulk_update_serializer

    def get_queryset(self):
        return range_list_filter('id', Assembly_Item.objects.all(),
                                 self.request)



#
#   Copyright (c) 2014-2021 eGauge Systems LLC
#	1644 Conestoga St, Suite 2
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
import logging
import os
import tempfile

from pathlib import Path

import sexpdata

from epic.base import Enum, format_part_number

from .error import Error
from .parts_cache import PartsCache

KICAD_FOOTPRINT = sexpdata.Symbol('footprint')
KICAD_PATH = sexpdata.Symbol('path')
KICAD_PROP = sexpdata.Symbol('property')
KICAD_PROP_EP = 'EP'				# eGauge part number
KICAD_PROP_FOOTPRINT = 'Footprint'
KICAD_PROP_REFDES = 'Reference'
KICAD_PROP_SHEET_FILE = 'Sheet file'
KICAD_SHEET = sexpdata.Symbol('sheet')
KICAD_SYMBOL = sexpdata.Symbol('symbol')
KICAD_SYMBOL_INSTANCES = sexpdata.Symbol('symbol_instances')
KICAD_UUID = sexpdata.Symbol('uuid')

log = logging.getLogger(__name__)

def _prefetch_symbol(symbol):
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_PROP:
            if sexp[1] == KICAD_PROP_EP:
                try:
                    part_number = int(sexp[2])
                except ValueError:
                    log.warning('Invalid EPIC part number "%s".', sexp[2])
                PartsCache.prefetch(part_number)

def _update_symbol(symbol, updated_syms):
    '''Check the footprint of a schematic symbol and update it if
    necessary.  SYMBOL is the s-expression of the symbol to update.
    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
    have been updated.  The value of each entry is the new footprint
    for that symbol.  Returns True if the symbol was updated, False
    otherwise.

    '''
    has_changed = False
    part_number = footprint_sexp = None
    refdes = ''
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_UUID:
            uuid = sexp[1].value()
        elif sexp[0] == KICAD_PROP:
            prop_name = sexp[1]
            if prop_name == KICAD_PROP_EP:
                try:
                    part_number = int(sexp[2])
                except ValueError:
                    log.warning('Invalid EPIC part number "%s".', sexp[2])
            elif prop_name == KICAD_PROP_FOOTPRINT:
                footprint_sexp = sexp
            elif prop_name == KICAD_PROP_REFDES:
                refdes = sexp[2] + ' '

    if part_number is not None:
        try:
            part = PartsCache.get(part_number)
        except Error:
            log.warning('Part %s does not exist.',
                        format_part_number(part_number))
            return False
        footprint = part.footprint
        if not footprint:
            footprint = ''
            if part.mounting not in [Enum.MOUNTING_CHASSIS, Enum.MOUNTING_FREE]:
                log.warning('Part %s has no footprint defined.',
                            format_part_number(part_number))
            return False

        if footprint_sexp is None:
            log.warning('Footprint property missing in symbol.')
            return False

        old_footprint = footprint_sexp[2]
        if old_footprint != footprint:
            footprint_sexp[2] = footprint
            if not old_footprint:
                log.info('%sfootprint set to %s.', refdes, footprint)
            else:
                log.warning('%sfootprint changed from %s to %s.',
                            refdes, old_footprint, footprint)
            updated_syms[uuid] = footprint
            has_changed = True
    return has_changed

def _update_symbol_instance(symbol, new_footprint):
    '''Update a symbol instance whose footprint has changed.  SYMBOL is
    the s-expression of the symbol to update.  NEW_FOOTPRINT is the
    name of footprint to set for the symbol.

    '''
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_FOOTPRINT:
            sexp[1] = new_footprint
            return
    symbol.append([KICAD_FOOTPRINT, new_footprint])

def _update_sheet(dirname, symbol, updated_syms):
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_PROP and sexp[1] == KICAD_PROP_SHEET_FILE:
            sheet_path = dirname / sexp[2]
            _update_schematic(sheet_path, updated_syms)

def _update_sheets_and_footprints(dirname, schematic, updated_syms):
    '''Update the schematic symbol footprints in a schematic in modern
    format (post KiCad v5) as needed. PATH is the filename path of the
    schematic, OUT is a temporary output file to write the updated
    schematic file to.  UPDATED_SYMS is a dictionary of all symbol
    UUIDs whose footprints have been updated.  The value of each entry
    is the new footprint for that symbol.  Returns True if the
    schematic file changed, False otherwise.

    '''
    # prefetch all EPIC parts:
    for sexp in schematic:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_SYMBOL:
            _prefetch_symbol(sexp)

    changed = False
    for sexp in schematic:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_SHEET:
            _update_sheet(dirname, sexp, updated_syms)
        elif sexp[0] == KICAD_SYMBOL:
            changed |= _update_symbol(sexp, updated_syms)
    return changed

def _update_symbol_instances(symbol, updated_syms):
    '''Check SYMBOL's UUID is in UPDATED_SYMS and, if so, update its
    footprint with the name given as the value of the entry in
    UPDATED_SYMS.  Returns True if the symbol was updated, False
    otherwise.

    '''
    changed = False
    for sexp in symbol:
        if not isinstance(sexp, list):
            continue
        if sexp[0] == KICAD_PATH:
            path = sexp[1].split('/')
            sym_path = path[-1]
            if sym_path in updated_syms:
                _update_symbol_instance(sexp, updated_syms[sym_path])
                changed = True
    return changed

def _update_sym_list(schematic, updated_syms):
    '''Update the footprints in the symbol-instances list of the top-level
    schematic of a KiCad v6 or newer schematic file.  DIRNAME is the
    directory containing the schematic and FILENAME is the filename
    within that directory.  HAS_BACKUP is True if the schematic file
    already has been backed up (due to an earlier change).
    UPDATED_SYMS is a dictionary of all symbol UUIDs whose footprints
    have been updated.  The value of each entry is the new footprint
    for that symbol.

    '''
    changed = False
    for sexp in schematic:
        if isinstance(sexp, list) and sexp[0] == KICAD_SYMBOL_INSTANCES:
            if _update_symbol_instances(sexp, updated_syms):
                changed = True
    return changed

def _update_schematic(path, updated_syms):
    '''Update schematic FILENAME with footprints from the EPIC database.
    UPDATED_SYMS is a dictionary into which are entered the UUIDs of
    all the symbols whose footprints have been updated.  The value of
    each entry is the new footprint for that symbol.  Returns True if
    the FILENAME was updated (and hence a backup file was created),
    False otherwise.

    '''
    dirname = path.parent
    with tempfile.NamedTemporaryFile(dir=dirname, mode='w',
                                     encoding='utf-8') as out, \
         open(path, encoding='utf-8') as schem_file:
        schematic = sexpdata.load(schem_file)
        has_changes = _update_sheets_and_footprints(dirname, schematic,
                                                    updated_syms)
        has_changes |= _update_sym_list(schematic, updated_syms)
        if has_changes:
            sexpdata.dump(schematic, out)
            os.rename(path, path.with_name(path.name +  '~'))
            os.link(out.name, path)
            log.info('%s updated.', path)
    return has_changes

def update_footprints(epic_api, filename):
    '''Update KiCad schematic FILENAME with the footprints from the EPIC
    database.  Any sheets within the schematic are also updated.  The
    original schematic files are backed up with a filename that has
    tilde appended.

    This program should not be called while KiCad has the schematic
    file open as KiCad may otherwise overwrite the changes made here.

    FILENAME may be a string or a pathlib.Path object.

    '''
    path = Path(filename)
    if path.suffix == '':
        path = path.with_suffix('.kicad_sch')

    PartsCache.set_epic(epic_api)

    updated_syms = {}
    _update_schematic(path, updated_syms)



#
#   Copyright (c) 2021 eGauge Systems LLC
#	1644 Conestoga St, Suite 2
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''KiCad EPIC scripts.'''
import argparse
import sys

from argparse import RawDescriptionHelpFormatter as RawDescHelpFormatter
from pathlib import Path

import logging

from . import schematics
from .bom import BOM
from .epic_client import EPICAPIClient
from .error import Error

def epic_bom(epic_url, auth):
    '''Command-line interface to translate a KiCAD XML file to a BOM.'''
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=RawDescHelpFormatter)
    parser.add_argument('input-file', help='Name of XML file '
                        'generated by KiCad\'s \"Generate BOM\" command.')
    parser.add_argument('output-file', help='Name of output file '
                        'that will contain the bill-of-materials (BOM) in '
                        'a comma-separated-values (CSV) file.')
    parser.add_argument('-d', '--debug', action='store_const',
                        const=logging.DEBUG, dest='log_level',
                        default=logging.INFO, help='Show debug output.')
    parser.add_argument('-f', '--force-overwrite', action='store_true',
                        help='Force overwriting of existing EPIC BOM '
                        '(assembly) part even if the part appears to '
                        'have been modified manually.')
    parser.add_argument('--flavor', help='Select PCB flavor for which to '
                        'generate BOM for.')
    parser.add_argument('-q', '--quiet', action='store_const',
                        const=logging.ERROR, dest='log_level',
                        help='Silence all output except for errors.')
    parser.add_argument('-p', '--prefer-vendor', nargs='+',
                        help='Like --with-vendor-pn, but only output vendor '
                        'part numbers for the specified vendor(s).')
    parser.add_argument('-V', '--with-vendor-pn',
                        help='Include vendor part numbers in the CSV file.',
                        action='store_true')
    args = parser.parse_args()

    logging.basicConfig()
    log = logging.getLogger()
    log.setLevel(args.log_level)

    include_vendor_pn = args.with_vendor_pn or args.prefer_vendor

    epic_api = EPICAPIClient(epic_url, auth)
    input_file = getattr(args, 'input-file')
    try:
        bom = BOM(epic_api, input_file, 'eGauge', flavor=args.flavor,
                  toolname='epic-bom')
        bom.save_as_epic_assembly(force_update=args.force_overwrite)
        bom.save_as_csv(filename=getattr(args, 'output-file'),
                        include_vendor_pn=include_vendor_pn,
                        preferred_vendors=args.prefer_vendor)
    except Error as e:
        log.error(e.args[0])
        sys.exit(1)

def epic_footprints(epic_url, auth):
    '''Command-line interface to update KiCad schematic files with the
    footprints from EPIC.

    '''
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=RawDescHelpFormatter)
    parser.add_argument('input-file', type=Path,
                        help='Name of KiCad schematic file.')
    parser.add_argument('-d', '--debug', action='store_const',
                        const=logging.DEBUG, dest='log_level',
                        default=logging.INFO, help='Show debug output.')
    parser.add_argument('-q', '--quiet', action='store_const',
                        const=logging.ERROR, dest='log_level',
                        help='Silence all output except for errors.')
    args = parser.parse_args()

    logging.basicConfig()
    log = logging.getLogger()
    log.setLevel(args.log_level)

    epic_api = EPICAPIClient(epic_url, auth)

    path = getattr(args, 'input-file')
    if path.suffix == '.sch':
        log.error('Cannot update old (pre v6) KiCad schematics files.')
        sys.exit(1)

    schematics.update_footprints(epic_api, path)



# Generated by Django 2.2.6 on 2019-10-03 22:43

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0008_auto_20180203_0553'),
    ]

    operations = [
        migrations.CreateModel(
            name='Stock_Cache_Entry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ts', models.DateTimeField(help_text='Date and time when this transaction was created.', verbose_name='Creation Time')),
            ],
            options={
                'ordering': ['-ts'],
            },
        ),
        migrations.CreateModel(
            name='Stock_Cache_Item',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('qty', models.IntegerField(help_text='The quantity of this part.')),
            ],
        ),
        migrations.AlterModelOptions(
            name='part',
            options={'ordering': ['id']},
        ),
        migrations.AlterField(
            model_name='vendor',
            name='name',
            field=models.CharField(max_length=31, unique=True, verbose_name='Vendor Name'),
        ),
        migrations.AlterIndexTogether(
            name='line_item',
            index_together={('txtn',), ('txtn', 'part')},
        ),
        migrations.AddIndex(
            model_name='delta',
            index=models.Index(fields=['txtn'], name='epic_delta_txtn_id_2ad9e1_idx'),
        ),
        migrations.AddIndex(
            model_name='delta',
            index=models.Index(fields=['warehouse'], name='epic_delta_warehou_9b26a4_idx'),
        ),
        migrations.AddIndex(
            model_name='vendor',
            index=models.Index(fields=['name'], name='epic_vendor_name_952e9d_idx'),
        ),
        migrations.AddField(
            model_name='stock_cache_item',
            name='part',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Part', verbose_name='Part #'),
        ),
        migrations.AddField(
            model_name='stock_cache_item',
            name='stock_cache_entry',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='epic.Stock_Cache_Entry'),
        ),
        migrations.AddField(
            model_name='stock_cache_item',
            name='warehouse',
            field=models.ForeignKey(help_text='The warehouse stocking this part.', on_delete=django.db.models.deletion.CASCADE, to='epic.Warehouse', verbose_name='Warehouse'),
        ),
        migrations.AddIndex(
            model_name='stock_cache_entry',
            index=models.Index(fields=['ts'], name='epic_stock__ts_25efd9_idx'),
        ),
        migrations.AddIndex(
            model_name='stock_cache_item',
            index=models.Index(fields=['stock_cache_entry'], name='epic_stock__stock_c_4ffb48_idx'),
        ),
    ]



#
#   Copyright (c) 2014-2015, 2018, 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
# pylint: disable=invalid-name
'''Unit tests.  Run, e.g., from epic-sample with:

	python3 manage.py test epic.tests
'''
import io
import os
import re

from datetime import timedelta, datetime
from decimal import Decimal

from django.test import TestCase
from django.utils import timezone
from django.contrib.auth import get_user_model
from django.contrib.auth import models as auth_models

import epic
from epic import perms
from epic.base import Enum
from epic.compat import reverse
from epic.lib import get_stock
from epic.models import Assembly_Item, Delta, Inventory, Line_Item, Order, \
    Part, Shipment, Vendor, Vendor_Part, Warehouse
from epic import urls

def mktime(time_string):
    dt = datetime.strptime(time_string, '%b %d %Y %I:%M%p')
    return timezone.make_aware(dt, timezone.utc)

class URL_Iterator:
    def __init__(self, urlpatterns):
        self.urlpattern_iterator = urlpatterns.__iter__()

    def __iter__(self):
        return self

    def __next__(self):
        # pylint: disable=protected-access
        try:
            pat = self.urlpattern_iterator.__next__() # Python 3
        except AttributeError:
            pat = self.urlpattern_iterator.next() # Python 2
        regex = re.compile(pat.pattern._regex)
        args = {}
        for key in regex.groupindex.keys():
            args[key] = 1
        url = reverse('epic:%s' % pat.name, kwargs=args)
        return url, pat.name

    def next(self):
        """For Python 2 compatibility."""
        return self.__next__()

def create_part(pk):
    return Part(pk=pk, mfg='mfg', mfg_pn='mfg_pn%d' % pk,
                mounting=Enum.MOUNTING_SMD, target_price=1, overage=0,
                spq=1, status=Enum.STATUS_ACTIVE, lead_time=9)

def create_warehouse(pk, name):
    return Warehouse(pk=pk, name=name)

def create_vendor(pk, name):
    return Vendor(pk=pk, name=name)

def create_vendor_part(part_id, vendor_id, vendor_pn, price):
    return Vendor_Part(part_id=part_id, vendor_id=vendor_id,
                       vendor_pn=vendor_pn, price=price)

def create_order(pk, vendor_id, warehouse_id):
    now = timezone.now()
    then = now + timedelta(days=30)
    return Order(pk=pk, ts=now, warehouse_id=warehouse_id,
                 expected_arrival_date=then,
                 status=Enum.TXTN_STATUS_OPEN, vendor_id=vendor_id)

def create_ship(pk, order_id, warehouse_id, tracking='1Z9AA10123456784',
                cost_freight=0, cost_other=0, cost_discount=0):
    return Shipment(pk=pk, ts=timezone.now(), tracking=tracking,
                    ordr_id=order_id, warehouse_id=warehouse_id,
                    cost_freight=cost_freight, cost_other=cost_other,
                    cost_discount=cost_discount)

def create_inventory(pk, warehouse_id):
    return Inventory(pk=pk, ts=timezone.now(), warehouse_id=warehouse_id)

def create_assembly_item(assy_id=None, comp_id=None, qty=1):
    return Assembly_Item(assy_id=assy_id, comp_id=comp_id, qty=qty)

def create_line_item(txtn_id, part_id, qty, line_cost, index):
    return Line_Item(txtn_id=txtn_id, part_id=part_id, qty=qty,
                     line_cost=line_cost, index=index)


class Stock_Cache_Tests(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.p = []	# parts
        cls.w = []	# warehouses
        cls.v = []	# vendors

        for i in range(10):
            p = create_part(i)
            p.save()
            cls.p.append(p)

        for i in range(3):
            w = create_warehouse(i, 'Warehouse%d' % i)
            w.save()
            cls.w.append(w)

        for i in range(5):
            v = create_vendor(i, 'Vendor%d' % i)
            v.save()
            cls.v.append(v)

    def assertInventory(self, inv, expected):
        for w in inv.warehouse:
            if w not in expected:
                self.fail('Warehouse %d unexpectedly in inventory.' % w)
            else:
                for p in inv.warehouse[w]:
                    if p not in expected[w]:
                        self.fail('Warehouse %d: part %d unexpectedly in '
                                  'inventory.' % (w, p))
                    else:
                        if inv.warehouse[w][p].qty != expected[w][p]:
                            self.fail('Warehouse %d, part %d: wrong qty %d; '
                                      'expected %d!' \
                                      % (w, p, inv.warehouse[w][p].qty,
                                         expected[w][p]))

                for p in expected[w]:
                    if p not in inv.warehouse[w]:
                        self.fail('Warehouse %d: part %d missing from '
                                  'inventory.' % (w, p))

        for w in expected:
            if w not in inv.warehouse:
                self.fail('Warehouse %d missing from inventory.' % w)

    def test_basic(self):
        inv0 = create_inventory(0, 0)
        inv0.ts = mktime('Jan 1 2010 4:00pm')
        inv0.save()
        inv0.finalize()

        d = Delta(part=self.__class__.p[0], is_absolute=True, adj=10, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[5], is_absolute=True, adj=20, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[3], is_absolute=True, adj=1000,
                  txtn=inv0, warehouse=self.__class__.w[2])
        d.save()

        inv = get_stock()
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})

        inv = get_stock(warehouses=[0])
        self.assertInventory(inv, {0: {0: 10, 5: 20}})

        inv = get_stock(warehouses=[1])
        self.assertInventory(inv, {})

        inv = get_stock(warehouses=[2])
        self.assertInventory(inv, {2: {3: 1000}})

        inv = get_stock(exclude_txtn_id=inv0.id)
        self.assertInventory(inv, {})

        inv = get_stock(as_of_ts=mktime('Jan 1 2010 3:59pm'))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime('Jan 1 2010 5:00pm'))
        self.assertInventory(inv, {})

    def test_shipment_timestamp_change(self):
        '''Test shipment that moves between timestamps.'''
        o = create_order(pk=0, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=1, order_id=o.id, warehouse_id=0)
        s.ts = mktime('Jan 1 2010 4:00pm')
        s.save()
        l = create_line_item(txtn_id=s.id, part_id=1, qty=2, line_cost='2',
                             index=1)
        l.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {0: {1: 2}})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

        s.ts = mktime('Jan 1 2017 5:00pm')
        s.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

        s.ts = mktime('Jan 1 2010 4:00pm')
        s.save()
        s.finalize()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {0: {1: 2}})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

    def test_inventory_timestamp_change(self):
        '''Test inventory that moves between timestamps.'''
        inv0 = create_inventory(0, 0)
        inv0.ts = mktime('Jan 1 2010 4:00pm')
        inv0.save()
        inv0.finalize()

        d = Delta(part=self.__class__.p[0], is_absolute=True, adj=10, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[5], is_absolute=True, adj=20, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[3], is_absolute=True, adj=1000,
                  txtn=inv0, warehouse=self.__class__.w[2])
        d.save()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

        inv0.ts = mktime('Jan 1 2017 5:00pm')
        inv0.save()
        inv0.finalize()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

        inv0.ts = mktime('Jan 1 2010 4:00pm')
        inv0.save()
        inv0.finalize()

        inv = get_stock(as_of_date=mktime('Jan 1 2015 5:00pm'))
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})
        inv = get_stock(as_of_date=mktime('Jan 1 2001 5:00pm'))
        self.assertInventory(inv, {})

    def test_warehouse_delete(self):
        '''Test warehouse delete effect on stock cache.'''
        item = create_assembly_item(assy_id=3, comp_id=1, qty=10)
        item.save()

        inv0 = create_inventory(pk=0, warehouse_id=2)
        inv0.ts = mktime('Jan 1 2010 4:00pm')
        inv0.save()
        inv0.finalize()

        d = Delta(part=self.__class__.p[0], is_absolute=True, adj=10, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[5], is_absolute=True, adj=20, txtn=inv0,
                  warehouse=self.__class__.w[0])
        d.save()
        d = Delta(part=self.__class__.p[3], is_absolute=True, adj=1000,
                  txtn=inv0, warehouse=self.__class__.w[2])
        d.save()

        inv = get_stock()
        self.assertInventory(inv, {0: {0: 10, 5: 20}, 2: {3: 1000}})

        Warehouse.objects.filter(pk=0).delete()

        inv = get_stock()
        self.assertInventory(inv, {2: {3: 1000}})

class Part_Method_Tests(TestCase):

    def test_is_orderable(self):
        p = Part(status=Enum.STATUS_PREVIEW)
        self.assertEqual(p.is_orderable(), False)

        p = Part(status=Enum.STATUS_ACTIVE)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Enum.STATUS_DEPRECATED)
        self.assertEqual(p.is_orderable(), True)

        p = Part(status=Enum.STATUS_OBSOLETE)
        self.assertEqual(p.is_orderable(), False)

    def test_equivalent_parts(self):
        p1 = create_part(1)
        p1.save()
        p2 = create_part(2)
        p2.save()

        self.assertEqual(p1.equivalent_parts(), [p1])

        p1.substitute = p2
        p1.save()
        p2.save()
        self.assertEqual(p1.equivalent_parts(), [p1, p2])
        self.assertEqual(p2.equivalent_parts(), [p1, p2])

        p3 = create_part(3)
        p3.substitute = p2
        p3.save()
        self.assertEqual(p1.equivalent_parts(), [p1, p2, p3])
        self.assertEqual(p2.equivalent_parts(), [p1, p2, p3])
        self.assertEqual(p3.equivalent_parts(), [p1, p2, p3])

        p1.substitute = None
        p1.save()
        self.assertEqual(p1.equivalent_parts(), [p1])
        self.assertEqual(p2.equivalent_parts(), [p2, p3])
        self.assertEqual(p3.equivalent_parts(), [p2, p3])

        p1.delete()
        p2.delete()
        p3.delete()

    def test_best_parts(self):
        p1 = create_part(1)
        p1.status = Enum.STATUS_PREFERRED
        p1.save()

        p2 = create_part(2)
        p2.status = Enum.STATUS_PREFERRED
        p2.substitute = p1
        p2.save()

        # p1 has lower ID so it should be preferred:
        self.assertEqual(p1.best_part(), p1)
        self.assertEqual(p2.best_part(), p1)

        p2.target_price = p1.target_price - 1
        p2.save()

        # p2 has lower price so it should be preferred:
        self.assertEqual(p1.best_part(), p2)
        self.assertEqual(p2.best_part(), p2)

        p2.status = Enum.STATUS_ACTIVE
        p2.save()
        # p1 is preferred, even though it's higher price
        self.assertEqual(p1.best_part(), p1)
        self.assertEqual(p2.best_part(), p1)

        p1.status = Enum.STATUS_ACTIVE
        p1.save()
        # now p2 should win again due to lower price:
        self.assertEqual(p1.best_part(), p2)
        self.assertEqual(p2.best_part(), p2)

    def test_avg_cost(self):
        part = []
        for i in range(3):
            p = create_part(i)
            if i == 1:
                p.substitute = part[0]
            p.save()
            part.append(p)

        for i in range(3):
            w = create_warehouse(i, 'Warehouse%d' % i)
            w.save()

        for i in range(5):
            v = create_vendor(i, 'Vendor%d' % i)
            v.save()

        # create shipment with:
        #	part 1 cost average = $1
        #	part 2 cost average = $0.001
        o = create_order(pk=0, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=1, order_id=o.id, warehouse_id=0)
        s.save()
        l = create_line_item(txtn_id=s.id, part_id=0, qty=1, line_cost='3',
                             index=1)
        l.save()
        l = create_line_item(txtn_id=s.id, part_id=1, qty=1, line_cost='1',
                             index=2)
        l.save()
        l = create_line_item(txtn_id=s.id, part_id=2, qty=1000, line_cost='1',
                             index=3)
        l.save()
        s.finalize()

        # Create shipment over a year ago.  It should not affect average prices
        o = create_order(pk=2, vendor_id=0, warehouse_id=0)
        o.save()
        o.finalize()
        s = create_ship(pk=3, order_id=o.id, warehouse_id=0)
        s.ts = timezone.now() - timedelta(days=365)
        s.save()
        l = create_line_item(txtn_id=s.id, part_id=1, qty=1, line_cost='1000',
                             index=1)
        l.save()
        l = create_line_item(txtn_id=s.id, part_id=2, qty=1, line_cost='1000',
                             index=2)
        l.save()
        s.finalize()

        p = Part.objects.get(pk=0)
        self.assertEqual(p.avg_cost(), Decimal('2.000'))

        p = Part.objects.get(pk=1)
        self.assertEqual(p.avg_cost(), Decimal('2.000'))

        p = Part.objects.get(pk=2)
        self.assertEqual(p.avg_cost(), Decimal('0.001'))

class Access_Tests(TestCase):
    @classmethod
    def setup_part(cls):
        p1 = create_part(1)
        p1.save()
        ai = create_assembly_item(assy_id=1, comp_id=2, qty=3)
        ai.refdes = 'R1,R2,R3'
        ai.save()
        p = create_vendor_part(1, 2, 'ASSY_HOUSE_VENDOR_PART_#1', 325.654321)
        p.save()

    @classmethod
    def setup_vendor(cls):
        v = create_vendor(1, 'Vendor #1')
        v.save()
        p = create_vendor_part(1, 1, 'VENDOR_PART_#1', 535.123456)
        p.save()

    @classmethod
    def setup_warehouse(cls):
        w = create_warehouse(1, 'Warehouse #1')
        w.save()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        perms.add_custom_permissions(auth_models)
        u = get_user_model().objects.create_user('test', 'test@magic.com',
                                                 'magic')
        u.user_permissions.set([perms.VIEW_PERM, perms.EDIT_PERM])

        p2 = create_part(2)
        p2.save()

        p3 = create_part(3)
        p3.save()

        w = create_warehouse(2, 'Assyhouse #1')
        w.save()
        v = create_vendor(2, 'Assyhouse #1')
        v.save()

        Access_Tests.setup_part()
        Access_Tests.setup_vendor()
        Access_Tests.setup_warehouse()

    # def test_redirect_to_https(self):
    #     url = reverse('epic:epic_index')
    #     response = self.client.get(url, secure=False)
    #     self.assertEqual(response.status_code, 301)
    #     self.assertIsNotNone(re.match('https://', response.url))

    def test_unauthenticated_accesses(self):
        response = self.client.get(reverse('epic:part_info') + '?pid=1',
                                   secure=True)
        self.assertEqual(response.status_code, 302)

        for url, _ in URL_Iterator(urls.urlpatterns):
            m = re.match(r'/epic/dal/', url)
            # auto-completion URLs respond with 403 (Forbidden)
            # others return redirect to login page:
            if m:
                expected = 403
            else:
                expected = 302
            response = self.client.get(url, secure=True)
            if response.status_code != expected:
                print('Unauthenticated access to %s returned %s, expected %s' %
                      (url, response.status_code, expected))
            self.assertEqual(response.status_code, expected)

    def test_authenticated_accesses(self):
        self.client.login(username='test', password='magic')

        for url, pattern_name in URL_Iterator(urls.urlpatterns):
            if re.match('order_', pattern_name):
                # order part 1 (an assembly) from vendor 'Assyhouse #1'
                o = create_order(1, 2, 1)
                o.save()
                l = create_line_item(1, 1, qty=13, line_cost='15.95', index=1)
                l.save()
            elif re.match('ship_', pattern_name):
                # order part 1(an assembly) from vendor 'Assyhouse #1'
                o = create_order(2, 2, 1)
                o.save()
                l = create_line_item(2, 1, qty=13, line_cost='15.95', index=1)
                l.save()
                s = create_ship(1, 2, 1, '1234,5678', 1.1, 2.2, 3.3)
                s.save()
                l = create_line_item(1, 1, qty=2, line_cost='2', index=1)
                l.save()
            elif re.match('warehouse_inventory_', pattern_name):
                i = create_inventory(1, 1)
                i.save()

            expect = 200
            expect_url = None
            post_action = []
            if pattern_name == 'search_results':
                url += '?q=1'
                # with a unique result, search redirects to the detail page:
                expect = 302
                expect_url = reverse('epic:part_detail', kwargs={'pk': 1})
            elif pattern_name == 'part_delete':
                expect = 302
                expect_url = reverse('epic:part_list')
                post_action.append(Access_Tests.setup_part)
            elif pattern_name == 'ship_delete':
                expect = 302
                expect_url = reverse('epic:ship_list')
            elif pattern_name == 'vendor_delete':
                expect = 302
                expect_url = reverse('epic:vendor_list')
                post_action.append(Access_Tests.setup_vendor)
            elif pattern_name == 'warehouse_delete':
                expect = 302
                expect_url = reverse('epic:warehouse_list')
                post_action.append(Access_Tests.setup_warehouse)
            elif pattern_name == 'warehouse_inventory_delete':
                expect = 302
                expect_url = reverse('epic:warehouse_detail', kwargs={'pk':1})
            elif pattern_name == 'order_delete':
                expect = 302
                expect_url = reverse('epic:order_list')
            elif pattern_name in ['datasheet_add_part', 'datasheet_detail',
                                  'datasheet_edit', 'datasheet_delete']:
                continue

            response = self.client.get(url, secure=True)

            # we can't only have one transaction with pk=1...
            Shipment.objects.all().delete()
            Order.objects.all().delete()
            Inventory.objects.all().delete()

            if response.status_code != expect:
                print('access to %s failed' % url)
                if hasattr(response, 'url'):
                    print(' redirect URL: %s' % response.url)
            self.assertEqual(response.status_code, expect)
            if expect_url is not None:
                url = response.url
                if url[-1] == '/':
                    url = url[:-1]
                self.assertEqual(url, expect_url)

            for action in post_action:
                action()

    def test_form_upload(self):
        self.client.login(username='test', password='magic')

        # Verify that we can post item formset to an order:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'line_item_set-TOTAL_FORMS': 12,
                                        'line_item_set-INITIAL_FORMS': 0,
                                        'line_item_set-MIN_NUM_FORMS': 0,
                                        'line_item_set-MAX_NUM_FORMS': 1000,
                                        'line_item_set-0-qty': 8,
                                        'line_item_set-0-part': 3,
                                        'line_item_set-0-vendor_pn': 'VPN #1',
                                        'line_item_set-0-part_cost': 1,
                                        'line_item_set-0-line_cost': 120,
                                        'line_item_set-1-qty': 18,
                                        'line_item_set-1-part': 2,
                                        'line_item_set-1-vendor_pn': 'VPN #2',
                                        'line_item_set-1-part_cost': 1,
                                        'line_item_set-1-line_cost': 121
                                    })

        # response.request['wsgi.input']._FakePayload__content.seek(0)
        # print('REQUEST',
        #       response.request['wsgi.input']._FakePayload__content.read()
        #       .decode('utf-8'))
        if response.status_code != 302:
            for name in response.context.keys():
                if re.match(r'[^_]+_form$', name):
                    print('%s.errors:' % name, response.context[name].errors)

        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 8)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 18)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post an item formset to a shipment:
        response = self.client.post('/epic/order/%d/add-shipment' % order.id,
                                    {
                                        'ts': '2018-08-06',
                                        'cost_freight': 10.0,
                                        'cost_other': 3.14,
                                        'cost_discount': 2.74,
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'note 2',
                                        'line_item_set-TOTAL_FORMS': 12,
                                        'line_item_set-INITIAL_FORMS': 0,
                                        'line_item_set-MIN_NUM_FORMS': 0,
                                        'line_item_set-MAX_NUM_FORMS': 1000,
                                        'line_item_set-0-qty': 7,
                                        'line_item_set-0-part': 3,
                                        'line_item_set-0-vendor_pn': 'VPN #1',
                                        'line_item_set-0-part_cost': 1,
                                        'line_item_set-0-line_cost': 120,
                                        'line_item_set-1-qty': 17,
                                        'line_item_set-1-part': 2,
                                        'line_item_set-1-vendor_pn': 'VPN #2',
                                        'line_item_set-1-part_cost': 1,
                                        'line_item_set-1-line_cost': 121
                                    })
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 7)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 120)
        self.assertEqual(line_items[1].qty, 17)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 121)

        # Verify that we can post a delta formset as an inventory:
        response = self.client.post('/epic/warehouse/1/add-inventory',
                                    {
                                        'ts': '2018-08-06 08:15 AM',
                                        'notes': 'Note #3',
                                        'delta_set-TOTAL_FORMS': 12,
                                        'delta_set-INITIAL_FORMS': 0,
                                        'delta_set-MIN_NUM_FORMS': 0,
                                        'delta_set-MAX_NUM_FORMS': 1000,
                                        'delta_set-0-adj': 6,
                                        'delta_set-0-part': 2,
                                        'delta_set-0-vendor_pn': 'VPN #2',
                                        'delta_set-0-part_cost': 1,
                                        'delta_set-0-line_cost': 120,
                                        'delta_set-1-adj': 16,
                                        'delta_set-1-part': 3,
                                        'delta_set-1-vendor_pn': 'VPN #3',
                                        'delta_set-1-part_cost': 1,
                                        'delta_set-1-line_cost': 121
                                    })
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 6)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 16)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_spreadsheet_upload(self):
        self.client.login(username='test', password='magic')
        # import pdb
        # pdb.set_trace()

        # Verify invalid column names yield the expected errors:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'spreadsheet': io.BytesIO(
                                            b'qty1,part1,extra,amount1\n'
                                            b'10,3,5,99\n'
                                            b'40,EP0002,,10\n'),
                                        'line_item_set-TOTAL_FORMS': 0,
                                        'line_item_set-INITIAL_FORMS': 0
                                    })
        self.assertFormError(response, 'order_form', 'spreadsheet',
                             ['Column qty missing.  It may also be called '
                              'one of: quantity, count.',
                              'Column part missing.  It may also be called '
                              'one of: pn, ep.',
                              'Column amount missing.  It may also be called '
                              'one of: cost.'])

        # Verify that we can post a CSV file to an order:
        response = self.client.post('/epic/order/add',
                                    {
                                        'ts': '2018-08-06',
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'This is a note',
                                        'spreadsheet': io.BytesIO(
                                            b'qty,part,extra,amount\n'
                                            b'10,3,5,99\n'
                                            b'40,EP0002,,10\n'),
                                        'line_item_set-TOTAL_FORMS': 0,
                                        'line_item_set-INITIAL_FORMS': 0
                                    })
        self.assertEqual(response.status_code, 302)
        order = Order.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=order)
        self.assertEqual(line_items[0].qty, 10)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 40)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post a CSV file to a shipment:
        response = self.client.post('/epic/order/%d/add-shipment' % order.id,
                                    {
                                        'ts': '2018-08-06',
                                        'cost_freight': 10.0,
                                        'cost_other': 3.14,
                                        'cost_discount': 2.74,
                                        'vendor': 1,
                                        'warehouse': 2,
                                        'expected_arrival_date': '2018-08-06',
                                        'notes': 'note 2',
                                        'spreadsheet': io.BytesIO(
                                            b'qty,part,extra,amount\n'
                                            b'5,3,5,99\n'
                                            b'10,EP0002,,10\n')
                                    })
        self.assertEqual(response.status_code, 302)
        ship = Shipment.objects.all()[0]
        line_items = Line_Item.objects.filter(txtn=ship)
        self.assertEqual(line_items[0].qty, 5)
        self.assertEqual(line_items[0].part.id, 3)
        self.assertEqual(line_items[0].line_cost, 99)
        self.assertEqual(line_items[1].qty, 10)
        self.assertEqual(line_items[1].part.id, 2)
        self.assertEqual(line_items[1].line_cost, 10)

        # Verify that we can post an Excel file as an inventory:
        excel_file = os.path.join(os.path.dirname(epic.__file__),
                                  'test-data', 'inventory.xlsx')
        with open(excel_file, 'rb') as f:
            response = self.client.post('/epic/warehouse/1/add-inventory',
                                        {
                                            'ts': '2018-08-06 08:15 AM',
                                            'spreadsheet': f
                                        })
        self.assertEqual(response.status_code, 302)
        inv = Inventory.objects.all()[0]
        line_items = Delta.objects.filter(txtn=inv)
        self.assertEqual(line_items[0].part.id, 2)
        self.assertEqual(line_items[0].is_absolute, True)
        self.assertEqual(line_items[0].adj, 1234)
        self.assertEqual(line_items[0].warehouse.id, 1)
        self.assertEqual(line_items[1].part.id, 3)
        self.assertEqual(line_items[1].is_absolute, True)
        self.assertEqual(line_items[1].adj, 5678)
        self.assertEqual(line_items[1].warehouse.id, 1)

    def test_order_status(self):
        '''Confirm that deleting a shipment for completed order will mark
        order as OPEN again.

        '''
        self.client.login(username='test', password='magic')
        # create order from Vendor #1:
        o = create_order(pk=4, vendor_id=1, warehouse_id=1)
        o.save()
        l = create_line_item(o.pk, 2, qty=1000, line_cost='256.35', index=1)
        l.save()

        # order should be open now
        self.assertEqual(o.status, Enum.TXTN_STATUS_OPEN)

        # create two shipments that should complete the order:
        s = create_ship(2, o.pk, 1, '1Z1345', 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 900, line_cost='230.05', index=1)
        l.save()

        s = create_ship(3, o.pk, 1, '1Z1356', 1.23, 2.34, 3.45)
        s.save()
        l = create_line_item(s.pk, 2, 100, line_cost='230.05', index=1)
        l.save()

        s.finalize()

        # order should be closed now:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Enum.TXTN_STATUS_CLOSED)

        # delete the first shipment:
        url = reverse('epic:ship_delete', kwargs={'pk':2})
        response = self.client.get(url, secure=True)
        self.assertEqual(response.status_code, 302)

        # order should be open again:
        o = Order.objects.get(pk=4)
        self.assertEqual(o.status, Enum.TXTN_STATUS_OPEN)



#
#   Copyright (c) 2021 eGauge Systems LLC
# 	1644 Conestoga St, Suite 2
# 	Boulder, CO 80301
# 	voice: 720-545-9767
# 	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
from django.urls import include, path

from rest_framework import routers

from . import views

router = routers.SimpleRouter()
router.register(r'part', views.PartViewSet, basename='part')
router.register(r'vendor', views.VendorViewSet, basename='vendor')
router.register(r'vendor_part', views.VendorPartViewSet,
                basename='vendor_part')
router.register(r'assembly_item', views.AssemblyItemViewSet,
                basename='assembly_view')

urlpatterns = [
    path('', include(router.urls))
]



#
#   Copyright (c) 2021 eGauge Systems LLC
#	1644 Conestoga St, Suite 2
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''This module provides helper functions outputting the BOM as a CSV
file.'''

from collections import OrderedDict
from types import SimpleNamespace

import io

from epic.base import Enum, format_part_number

from .error import Error
from .parts_cache import PartsCache

def _bom_sort(bom):
    '''Sort a dictionary of components by part number and return the
    resulting list.

    '''
    bom_list = []
    for component in bom.values():
        bom_list.append(component)
    bom_list.sort(key=lambda x: x[0].part_id)
    return bom_list

def _quote(string):
    '''CSV quote a string.  DigiKey doesn't handle commas correctly even
    if they're inside double quotes so we also replace commas with
    colons.

    '''
    string = string.replace('"', '\"')
    string = string.replace(',', ':')
    return '"' + string + '"'

def _is_orderable(vp):
    '''A vendor part VP is orderable only if its status indicates its
    orderable and also it corresponding part is also orderable.

    '''
    if vp.status not in Enum.STATUS_ORDERABLE:
        return False
    part = PartsCache.get(vp.part)
    if part.status not in Enum.STATUS_ORDERABLE:
        return False
    return True

def _best_vendor_parts(epic_api, component, vendors, preferred_vendors):
    '''Find the best vendor parts for a given COMPONENT.  Orderable parts
    are better than deprecated or obsolete parts.  Lower cost parts
    are better than higher cost parts.  Note that all vendor parts
    from equivalent parts (substitutes) are considered, not just the
    vendor parts for the part specified directly by COMPONENT.

    VENDORS must be a dictionary of vendors to consider, indexed by
    vendor id.

    If PREFERRED_VENDORS is None, a dictionary of best parts for all
    VENDORS is returned, indexed by vendor.  Otherwise, the returned
    dictionary is guaranteed to have an entry for each vendor in
    VENDORS.  The entry may have a value of None if a particular
    VENDOR does not carry the part.

    '''
    ret = OrderedDict()
    if preferred_vendors is not None:
        for vendor in vendors.values():
            ret[vendor.id] = None

    parts = ['%d' % part_number for part_number in component.part.equivalents]

    reply = epic_api.get('vendor_part/?part_id=%s' % ','.join(parts))
    if not reply:
        return ret

    for vp in reply:
        part = SimpleNamespace(**vp)
        vendor = vendors.get(part.vendor)
        if vendor is None:
            continue
        if vendor.id in vendors:
            if vendor.id in ret and ret[vendor.id] is not None:
                if _is_orderable(ret[vendor.id]) and not _is_orderable(part):
                    continue
                if not _is_orderable(ret[vendor.id]) and _is_orderable(part) \
                   or ret[vendor.id].price < part.price:
                    continue
            ret[vendor.id] = part
    return ret

def _output_list(outfile, epic_api, bom_list, with_vendor_pn=False,
                 vendors=None, preferred_vendors=None):
    '''Output the components in BOM_LIST in CSV format to file OUTFILE.
    EPIC_API is the EPIC-API-client to use for fetching EPIC info.
    WITH_VENDOR_PN must be True if vendor part numbers should be
    output to the CSV file.  VENDORS is a dictionary of vendors,
    indexed by vendor-id.  PREFERRED_VENDORS should by a value other
    than None if vendor part-number for each of the VENDORS must be
    output.

    '''
    for components in bom_list:
        c = components[0]
        if c.part:
            for sub_id in c.part.equivalents:
                PartsCache.prefetch(sub_id)

    for components in bom_list:
        qty = len(components)
        refdes = ','.join(sorted([c.refdes for c in components]))
        c = components[0]

        substitutes = ''
        if c.part:
            sep = ''
            for sub_id in c.part.equivalents:
                sub = PartsCache.get(sub_id)
                if sub_id == c.part_id:
                    continue
                substitutes += sep
                substitutes += '%s (%s %s)' % (format_part_number(sub.id),
                                               sub.mfg, sub.mfg_pn)
                sep = ', '

        row = '%s,%s,%s,%s,%s,%s,%s,%s' % \
              (qty, c.value, c.footprint, format_part_number(c.part_id),
               c.mfg, _quote(c.mfg_pn), _quote(substitutes), _quote(refdes))

        if with_vendor_pn:
            vps = _best_vendor_parts(epic_api, c, vendors, preferred_vendors)
            if preferred_vendors is None:
                for vid, part in vps.items():
                    row += ',%s,%s' % (vendors[vid].name,
                                       part.vendor_pn if part else 'n/a')
            else:
                for part in vps.values():
                    row += ',%s' % (part.vendor_pn if part else 'n/a')
        print(row, file=outfile)

def write(filename, epic_api, bom,
          include_vendor_pn=False, preferred_vendors=None):
    '''Output the BOM to a file named FILENAME in CSV format.  EPIC_API is
    the EPIC-API-client to use for fetching EPIC info.
    INCLUDE_VENDOR_PN should be True if vendor part information should
    be output.  PREFERRED_VENDORS can be set to a list to output
    vendor part numbers only for the specified vendors.

    '''
    reply = epic_api.get('vendor/')
    if not reply:
        raise Error('Unable to get vendor list.')
    vendors = OrderedDict()
    for v in reply:
        vendor = SimpleNamespace(**v)
        if preferred_vendors is None:
            vendors[vendor.id] = vendor
        else:
            for preferred in preferred_vendors:
                if preferred.lower() in vendor.name.lower():
                    vendors[vendor.id] = vendor

    with io.open(filename, 'w', encoding='utf-8') as out:
        hdr = (('Qty,Value,Footprint,%s PN,Mfg,Mfg PN,' % bom.manufacturer)
               + 'Approved Substitutes,Refdes')
        if include_vendor_pn:
            if preferred_vendors is None:
                hdr += u',Vendor,Vendor PN...'
            else:
                for vendor in vendors.values():
                    hdr += u',%s' % _quote(vendor.name + ' PN')
        print(hdr, file=out)

        comp_list = _bom_sort(bom.comps)
        _output_list(out, epic_api, comp_list, include_vendor_pn,
                     vendors, preferred_vendors)

        dnp_list = _bom_sort(bom.dnps)
        if dnp_list:
            print('\nDO NOT PLACE parts:', file=out)
            _output_list(out, epic_api, dnp_list, include_vendor_pn,
                         vendors, preferred_vendors)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('epic', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='part',
            name='last_bom_mod_name',
            field=models.CharField(default=b'', help_text=b'Name of entity which last modified this part.', max_length=31, blank=True),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='part',
            name='last_bom_mod_type',
            field=models.IntegerField(default=0, choices=[(0, b'user'), (1, b'tool')]),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='vendor_part',
            name='status',
            field=models.IntegerField(default=1, help_text=b'The life-time status of this vendor part.', verbose_name=b'Life-time Status', choices=[(0, b'preview'), (1, b'active'), (2, b'deprecated'), (3, b'obsolete')]),
            preserve_default=True,
        ),
    ]



#
#   Copyright (c) 2014-2021 eGauge Systems LLC
#	1644 Conestoga St, Suite 2
#	Boulder, CO 80301
#	voice: 720-545-9767
#	email: davidm@egauge.net
#
#   All rights reserved.
#
#   This code is the property of eGauge Systems LLC and may not be
#   copied, modified, or disclosed without any prior and written
#   permission from eGauge Systems LLC.
#
'''This module handles parsing of KiCAD BOM files (XML files) and
translating them to an EPIC assembly and/or CSV output file.

'''

from types import SimpleNamespace

import logging
import re
import xml.etree.ElementTree as ET

from pathlib import Path

from egauge import webapi

from epic.base import Enum, format_part_number, strchoice

from . import csv
from .error import Error
from .parts_cache import PartsCache

EP_PAT = re.compile(r'EP(-(.*))?')
INST_PAT = re.compile(r'Installation(-(.*))?')
VALUE_PAT = re.compile(r'Value(-(.*))?')

log = logging.getLogger(__name__)

def _bom_append(bom, component):
    if component.part_id not in bom:
        bom[component.part_id] = []
    bom[component.part_id].append(component)

def _update(res, field, desired_flavor, detected_flavors):
    '''Process object RES by matching FIELD.attrib['name'] against
    RES.pattern.  If there is a match M, M.group(2) must evaluate to
    the name of the flavor of the field.  If this flavor matches the
    specified flavor, we have an exact match and RES.value is set to
    field.text, RES.name is set to field.attrib['name'].  If the
    flavor of the field is empty and RES.value is None we have a
    default match and RES is updated like for an exact match.
    DETECTED_FLAVOR is a set of flavors found.

    '''
    m = res.pattern.match(field.attrib['name'])
    if m is None:
        return

    this_flavor = None
    if m.lastindex is not None:
        this_flavor = m.group(2)
        detected_flavors |= {this_flavor} # update set of all detected flavors

    if this_flavor == desired_flavor \
       or (this_flavor is None and res.value is None):
        res.name = field.attrib['name']
        res.value = field.text

class Component:
    # dictionary of components that first referenced a particular best part:
    first_reference = {}

    def __init__(self, part_id, refdes, value, footprint):
        self.part_id = part_id
        self.part = None
        self.refdes = refdes
        self.value = value
        self.footprint = footprint
        self.mfg = None
        self.mfg_pn = None
        PartsCache.prefetch(part_id)

    def load_from_epic(self):
        self.part = PartsCache.get(self.part_id)

        if self.value != self.part.val:
            log.warning('%s has value "%s" but part %s has value "%s"',
                        self.refdes, self.value, self, self.part.val)

        if self.footprint != self.part.footprint:
            if self.footprint:
                msg = 'changed from %s' % self.footprint
            else:
                msg = 'set'
            log.warning('%s footprint should be %s to %s.',
                        self.refdes, msg, self.part.footprint)

        best_part = self.part.best_part
        first_ref = self.__class__.first_reference.get(best_part)
        if first_ref is None:
            self.__class__.first_reference[best_part] = self
        else:
            if self.part_id != first_ref.part_id:
                log.warning('%s uses %s instead of equivalent %s used by %s',
                            self.refdes, self, first_ref, first_ref.refdes)

        self.mfg = self.part.mfg
        self.mfg_pn = self.part.mfg_pn

    def __str__(self):
        return format_part_number(self.part_id)

class BOM:
    '''A BOM object is a bill-of-material for a given PCB.'''

    def __init__(self, epic_api, xml_filename, manufacturer, flavor=None,
                 toolname='kicad-to-epic-bom'):
        '''Create a BOM object that represents the KiCad intermediate XML
        netlist stored in file XML_FILENAME.

        EPIC_API must be an instance of EPICAPIClient which provides
        access to the EPIC JSON API.

        MANUFACTURER is the name to use as the creator of the board.
        This is typically a short version of the company that designed
        the board.  For example, "eGauge" might be used by "eGauge
        Systems LLC".

        A single Eeschema schematic may define multiple flavors
        (versions) of a PCB.  Each flavor results in its own BOM.  The
        flavor to be used is selected with argument FLAVOR.  If this
        argument is not set, the default ("unflavored") BOM is
        generated.

        The part number and the installation of that part is
        determined based on various fields of the schematic component.
        Specifically, the following fields are used:

	Field Name:	Purpose:
	------------	---------------------------------------------
	EP[-FLAVOR]	Specifies the EPIC part number to use for the
			component.  If -FLAVOR is specified, the field
			specifies the part number to use only for that
			FLAVOR.

	Installation[-FLAVOR]
			If the value of this field is "DNP" (Do Not Place),
			then the component is omitted from the BOM.
			In the CSV output file, do-not-place components
			are listed separately at the end of the file.
			If -FLAVOR is specified, the field specifies the
			installation-type of the component only for that
			FLAVOR.

        Flavored field-names take precedence over unflavored fields.
        For example, if a component specified the fields:

		EP	123
		EP-lite	42

        then part number 42 would be used for flavor "lite" but 123
        would be used in all other cases.

        '''
        self.manufacturer = manufacturer
        self.toolname = toolname
        self.schematic_name = 'unknown'
        self.revision = ''
        self.sources = []
        self.comps = {}		# dictionary of components in the BOM
        self.dnps = {}		# dictionary of do-not-place components
        self.epic_api = epic_api
        PartsCache.set_epic(epic_api)

        self.flavor = flavor

        try:
            xml = ET.parse(xml_filename)
        except ET.ParseError as _:
            raise Error('Input file is not a valid XML file.', xml_filename)

        design = xml.find('design')
        if design is not None:
            source = design.findtext('source')
            if source is not None:
                path = Path(source)
                self.schematic_name = path.with_suffix('').name
            sheet = design.find('sheet')
            for sheet in design.iter('sheet'):
                title_block = sheet.find('title_block')
                if title_block is not None:
                    if self.revision == '':
                        rev = title_block.find('rev')
                        if rev and rev.text:
                            self.revision = '-rev' + rev.text.lstrip().rstrip()
                    self.sources.append(title_block.findtext('source'))

        detected_flavors = set()
        for comp in xml.find('components'):
            refdes = comp.attrib.get('ref')

            footprint = comp.findtext('footprint')

            part = SimpleNamespace(name=None, value=None, pattern=EP_PAT)
            inst = SimpleNamespace(name=None, value=None, pattern=INST_PAT)
            value = SimpleNamespace(name=None, value=None, pattern=VALUE_PAT)
            fields = comp.find('fields')
            if fields is not None:
                for field in fields:
                    _update(part, field, flavor, detected_flavors)
                    _update(inst, field, flavor, detected_flavors)
                    _update(value, field, flavor, detected_flavors)

            do_not_place = (inst.value == 'DNP')

            if not part.value:
                if not do_not_place:
                    log.warning('%s skipped due to missing EPIC part number'
                                '(field "EP")', refdes)
                continue

            if value.value is None:
                value.value = comp.findtext('value', default='n/a')

            if do_not_place:
                log.info('%s marked as do-not-place ("%s=DNP")',
                         refdes, inst.name)

            try:
                part_id = int(part.value)
            except ValueError:
                log.warning('%s has invalid EPIC part number "%s"',
                            refdes, part.value)
                continue

            c = Component(part_id, refdes, value.value, footprint)

            if do_not_place:
                _bom_append(self.dnps, c)
            else:
                _bom_append(self.comps, c)

        for l in self.comps.values():
            for c in l:
                c.load_from_epic()
        for l in self.dnps.values():
            for c in l:
                c.load_from_epic()

        if flavor is not None and flavor not in detected_flavors:
            raise Error('Flavor does not exist', flavor, detected_flavors)

        if detected_flavors:
            if not detected_flavors:
                detected_flavors_str = 'no flavors'
            else:
                if len(detected_flavors) > 1:
                    detected_flavors_str = 'flavors: '
                else:
                    detected_flavors_str = 'flavor: '
                detected_flavors_str += ', '.join(sorted(detected_flavors))
            log.info('this schematic defines %s', detected_flavors_str)

    def save_as_epic_assembly(self, force_update=False):
        '''Save the BOM as an EPIC assembly.  If an EPIC assembly with the
        same part-number already exists, it is updated unless the
        assembly indicates that it was last updated by a different
        tool or an interactive EPIC user.  This is dected based on the
        last update type and the toolname set when the BOM object was
        created (see argument TOOLNAME). If the assembly was last
        edited by a different tool or interactive user, a kicad.Error
        exception is raised, unless FORCE_UPDATE is True.

        If an EPIC assembly item is created, its manufacturer is set
        to the MANUFACTURER specified when creating the BOM object and
        its part number will have the form bom:SCHEMATIC_NAME[-FLAVOR],
        where SCHEMATIC_NAME is the name of the schematic and FLAVOR is
        the name of the selected flavor (if any).

        '''
        assembly_name = 'bom:' + self.schematic_name + self.revision
        if self.flavor is not None:
            assembly_name += '-' + self.flavor

        # see if the assembly exists already:
        old_assy = None
        try:
            reply = self.epic_api.get('part/?mfg=%s&mfg_pn=%s' \
                                      % (self.manufacturer, assembly_name))
            if reply:
                old_assy = SimpleNamespace(**reply[0])
        except webapi.Error:
            pass

        if old_assy:
            if old_assy.last_bom_mod_type != Enum.LAST_MOD_TYPE_TOOL \
               or old_assy.last_bom_mod_name != self.toolname:
                last_editor = '%s %s' % (strchoice(Enum.LAST_MOD_CHOICES,
                                                   old_assy.last_bom_mod_type),
                                         old_assy.last_bom_mod_name)
                if force_update:
                    log.info('overwriting part %s %s last modified by %s',
                             self.manufacturer, assembly_name, last_editor)
                else:
                    raise Error('Refusing to overwrite part last modified '
                                'by %s %s.' % (last_editor,
                                               format_part_number(old_assy.id)))

        desc = 'BOM for %s' % self.schematic_name
        if self.flavor is not None:
            desc += (', ' + self.flavor + ' flavor')
        assembly_part = \
            SimpleNamespace(descr=desc,
                            mfg=self.manufacturer,
                            mfg_pn=assembly_name,
                            mounting=Enum.MOUNTING_CHASSIS,
                            target_price=1000,
                            overage=1, spq=1, lead_time=4,
                            status=Enum.STATUS_PREVIEW,
                            last_bom_mod_type=Enum.LAST_MOD_TYPE_TOOL,
                            last_bom_mod_name=self.toolname)

        try:
            if old_assy:
                reply = self.epic_api.put('part/%d/' % old_assy.id,
                                          assembly_part)
            else:
                reply = self.epic_api.post('part/', assembly_part)
            assembly_part = SimpleNamespace(**reply)
        except webapi.Error as e:
            raise Error('Failed to create assembly part.', assembly_part) \
                        from e

        # create assembly-items for the components in the BOM:
        assy_items = []
        for components in self.comps.values():
            comp = components[0]
            refdes = ','.join(sorted([c.refdes for c in components]))
            assy_item = SimpleNamespace(assy=assembly_part.id,
                                        comp=comp.part_id,
                                        qty=len(components),
                                        refdes=refdes)
            assy_items.append(assy_item.__dict__)
        try:
            reply = self.epic_api.post('assembly_item/', assy_items)
        except webapi.Error as e:
            raise Error('Failed to save assembly items.', assy_items) from e

        log.info('%s part %s (%s %s) with %d components',
                 'updated' if old_assy else 'created',
                 format_part_number(assembly_part.id),
                 assembly_part.mfg, assembly_part.mfg_pn, len(self.comps))

    def save_as_csv(self, filename, **kwargs):
        '''Save the BOM as a CSV text file.  See epic.kicad.csv.write() for a
        description of available parameters.

        '''
        csv.write(filename, self.epic_api, self, **kwargs)
